# futureserver.py
# future
from __future__ import annotations

# stdlib
from collections import OrderedDict
from collections.abc import Callable
from datetime import MINYEAR
from datetime import datetime
from datetime import timezone
from functools import partial
import hashlib
import logging
import os
from pathlib import Path
import threading
from time import sleep
import traceback
from typing import Any
from typing import TypeVar
from typing import cast

# third party
from nacl.signing import SigningKey

# relative
from .. import __version__
from ..abstract_server import AbstractServer
from ..abstract_server import ServerSideType
from ..abstract_server import ServerType
from ..client.api import SignedSyftAPICall
from ..client.api import SyftAPI
from ..client.api import SyftAPICall
from ..client.api import SyftAPIData
from ..client.api import debox_signed_syftapicall_response
from ..client.client import SyftClient
from ..deployment_type import DeploymentType
from ..protocol.data_protocol import PROTOCOL_TYPE
from ..protocol.data_protocol import get_data_protocol
from ..service.action.action_object import Action
from ..service.action.action_object import ActionObject
from ..service.code.user_code_stash import UserCodeStash
from ..service.context import AuthedServiceContext
from ..service.context import ServerServiceContext
from ..service.context import UnauthedServiceContext
from ..service.context import UserLoginCredentials
from ..service.job.job_stash import Job
from ..service.job.job_stash import JobStash
from ..service.job.job_stash import JobStatus
from ..service.job.job_stash import JobType
from ..service.metadata.server_metadata import ServerMetadata
from ..service.network.utils import PeerHealthCheckTask
from ..service.notifier.notifier_service import NotifierService
from ..service.output.output_service import OutputStash
from ..service.queue.base_queue import AbstractMessageHandler
from ..service.queue.base_queue import QueueConsumer
from ..service.queue.base_queue import QueueProducer
from ..service.queue.queue import APICallMessageHandler
from ..service.queue.queue import ConsumerType
from ..service.queue.queue import QueueManager
from ..service.queue.queue_stash import APIEndpointQueueItem
from ..service.queue.queue_stash import ActionQueueItem
from ..service.queue.queue_stash import QueueItem
from ..service.queue.queue_stash import QueueStash
from ..service.queue.zmq_client import QueueConfig
from ..service.queue.zmq_client import ZMQClientConfig
from ..service.queue.zmq_client import ZMQQueueConfig
from ..service.response import SyftError
from ..service.response import SyftSuccess
from ..service.service import AbstractService
from ..service.service import ServiceConfigRegistry
from ..service.service import UserServiceConfigRegistry
from ..service.settings.settings import ServerSettings
from ..service.settings.settings import ServerSettingsUpdate
from ..service.user.user import UserView
from ..service.user.user_roles import ServiceRole
from ..service.user.utils import create_root_admin_if_not_exists
from ..service.worker.utils import DEFAULT_WORKER_IMAGE_TAG
from ..service.worker.utils import DEFAULT_WORKER_POOL_NAME
from ..service.worker.utils import create_default_image
from ..service.worker.worker_pool import WorkerPool
from ..service.worker.worker_pool_service import SyftWorkerPoolService
from ..service.worker.worker_pool_stash import SyftWorkerPoolStash
from ..service.worker.worker_stash import WorkerStash
from ..store.blob_storage import BlobStorageConfig
from ..store.blob_storage.on_disk import OnDiskBlobStorageClientConfig
from ..store.blob_storage.on_disk import OnDiskBlobStorageConfig
from ..store.blob_storage.seaweedfs import SeaweedFSBlobDeposit
from ..store.db.db import DBConfig
from ..store.db.db import DBManager
from ..store.db.postgres import PostgresDBConfig
from ..store.db.postgres import PostgresDBManager
from ..store.db.sqlite import SQLiteDBConfig
from ..store.db.sqlite import SQLiteDBManager
from ..store.db.stash import ObjectStash
from ..store.document_store_errors import NotFoundException
from ..store.document_store_errors import StashException
from ..store.linked_obj import LinkedObject
from ..types.datetime import DATETIME_FORMAT
from ..types.errors import SyftException
from ..types.result import Result
from ..types.result import as_result
from ..types.syft_metaclass import Empty
from ..types.syft_object import Context
from ..types.syft_object import PartialSyftObject
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import SyftObject
from ..types.uid import UID
from ..util.experimental_flags import flags
from ..util.telemetry import instrument
from ..util.util import get_dev_mode
from ..util.util import get_env
from ..util.util import get_queue_address
from ..util.util import random_name
from ..util.util import thread_ident
from .credentials import SyftSigningKey
from .credentials import SyftVerifyKey
from .env import get_default_root_email
from .env import get_default_root_password
from .env import get_default_root_username
from .env import get_default_worker_image
from .env import get_default_worker_pool_name
from .env import get_default_worker_pool_pod_annotations
from .env import get_default_worker_pool_pod_labels
from .env import get_private_key_env
from .env import get_server_uid_env
from .env import get_syft_worker_uid
from .env import in_kubernetes
from .service_registry import ServiceRegistry
from .utils import get_named_server_uid
from .utils import get_temp_dir_for_server
from .utils import remove_temp_dir_for_server
from .worker_settings import WorkerSettings

logger = logging.getLogger(__name__)

SyftT = TypeVar("SyftT", bound=SyftObject)

# if user code needs to be serded and its not available we can call this to refresh
# the code for a specific server UID and thread
CODE_RELOADER: dict[int, Callable] = {}


def get_default_worker_pool_count(server: Server) -> int:
    return int(
        get_env(
            "DEFAULT_WORKER_POOL_COUNT", server.queue_config.client_config.n_consumers
        )
    )


signing_key_env = get_private_key_env()
server_uid_env = get_server_uid_env()

default_root_email = get_default_root_email()
default_root_username = get_default_root_username()
default_root_password = get_default_root_password()


class AuthServerContextRegistry:
    __server_context_registry__: dict[str, ServerServiceContext] = OrderedDict()

    @classmethod
    def set_server_context(
        cls,
        server_uid: UID | str,
        context: ServerServiceContext,
        user_verify_key: SyftVerifyKey | str,
    ) -> None:
        if isinstance(server_uid, str):
            server_uid = UID.from_string(server_uid)

        if isinstance(user_verify_key, str):
            user_verify_key = SyftVerifyKey.from_string(user_verify_key)

        key = cls._get_key(server_uid=server_uid, user_verify_key=user_verify_key)

        cls.__server_context_registry__[key] = context

    @staticmethod
    def _get_key(server_uid: UID, user_verify_key: SyftVerifyKey) -> str:
        return "-".join(str(x) for x in (server_uid, user_verify_key))

    @classmethod
    def auth_context_for_user(
        cls,
        server_uid: UID,
        user_verify_key: SyftVerifyKey,
    ) -> AuthedServiceContext | None:
        key = cls._get_key(server_uid=server_uid, user_verify_key=user_verify_key)
        return cls.__server_context_registry__.get(key)


class Server(AbstractServer):
    signing_key: SyftSigningKey | None
    required_signed_calls: bool = True
    packages: str

    def __init__(
        self,
        *,  # Trasterisk
        name: str | None = None,
        id: UID | None = None,
        signing_key: SyftSigningKey | SigningKey | None = None,
        db_config: DBConfig | None = None,
        root_email: str | None = default_root_email,
        root_username: str | None = default_root_username,
        root_password: str | None = default_root_password,
        processes: int = 0,
        is_subprocess: bool = False,
        server_type: str | ServerType = ServerType.DATASITE,
        deployment_type: str | DeploymentType = "remote",
        reset: bool = False,
        blob_storage_config: BlobStorageConfig | None = None,
        queue_config: QueueConfig | None = None,
        queue_port: int | None = None,
        n_consumers: int = 0,
        create_producer: bool = False,
        thread_workers: bool = False,
        server_side_type: str | ServerSideType = ServerSideType.HIGH_SIDE,
        enable_warnings: bool = False,
        dev_mode: bool = False,
        migrate: bool = False,
        in_memory_workers: bool = True,
        log_level: int | None = None,
        smtp_username: str | None = None,
        smtp_password: str | None = None,
        email_sender: str | None = None,
        smtp_port: int | None = None,
        smtp_host: str | None = None,
        association_request_auto_approval: bool = False,
        background_tasks: bool = False,
        consumer_type: ConsumerType | None = None,
        db_url: str | None = None,
    ):
        # 🟡 TODO 22: change our ENV variable format and default init args to make this
        # less horrible or add some convenience functions
        self.dev_mode = dev_mode or get_dev_mode()
        self.id = UID.from_string(server_uid_env) if server_uid_env else (id or UID())
        self.packages = ""
        self.processes = processes
        self.is_subprocess = is_subprocess
        self.name = name or random_name()
        self.enable_warnings = enable_warnings
        self.in_memory_workers = in_memory_workers
        self.server_type = ServerType(server_type)
        self.server_side_type = ServerSideType(server_side_type)
        self.client_cache: dict = {}
        self.peer_client_cache: dict = {}
        self._settings = None

        if isinstance(server_type, str):
            server_type = ServerType(server_type)

        self.server_type = server_type

        if isinstance(deployment_type, str):
            deployment_type = DeploymentType(deployment_type)
        self.deployment_type = deployment_type

        # do this after we set the deployment type
        self.set_log_level(log_level)

        if isinstance(server_side_type, str):
            server_side_type = ServerSideType(server_side_type)
        self.server_side_type = server_side_type

        skey = None
        if signing_key_env:
            skey = SyftSigningKey.from_string(signing_key_env)
        elif isinstance(signing_key, SigningKey):
            skey = SyftSigningKey(signing_key=signing_key)
        else:
            skey = signing_key
        self.signing_key = skey or SyftSigningKey.generate()

        logger.critical(
            f"Hash of the signing key '{self.signing_key.deterministic_hash()[:5]}...'"
        )

        self.association_request_auto_approval = association_request_auto_approval

        consumer_type = (
            consumer_type or ConsumerType.Thread
            if thread_workers
            else ConsumerType.Process
        )
        self.queue_config = self.create_queue_config(
            n_consumers=n_consumers,
            create_producer=create_producer,
            consumer_type=consumer_type,
            queue_port=queue_port,
            queue_config=queue_config,
        )

        # must call before initializing stores
        if reset:
            self.remove_temp_dir()

        db_config = DBConfig.from_connection_string(db_url) if db_url else db_config

        if db_config is None:
            db_config = SQLiteDBConfig(
                filename=f"{self.id}_json.db",
                path=self.get_temp_dir("db"),
            )

        self.db_config = db_config

        self.db = self.init_stores(db_config=self.db_config)

        # construct services only after init stores
        self.services: ServiceRegistry = ServiceRegistry.for_server(self)
        self.db.init_tables(reset=reset)
        self.action_store = self.services.action.stash

        create_root_admin_if_not_exists(
            name=root_username,
            email=root_email,
            password=root_password,  # nosec
            server=self,
        )

        NotifierService.init_notifier(
            server=self,
            email_password=smtp_password,
            email_username=smtp_username,
            email_sender=email_sender,
            smtp_port=smtp_port,
            smtp_host=smtp_host,
        ).unwrap()

        self.post_init()

        if migrate:
            self.find_and_migrate_data()
        else:
            self.find_and_migrate_data([ServerSettings])

        self.create_initial_settings(admin_email=root_email).unwrap()

        self.init_blob_storage(config=blob_storage_config)

        # Migrate data before any operation on db

        # first migrate, for backwards compatibility
        self.init_queue_manager(queue_config=self.queue_config)

        context = AuthedServiceContext(
            server=self,
            credentials=self.verify_key,
            role=ServiceRole.ADMIN,
        )

        self.peer_health_manager: PeerHealthCheckTask | None = None
        if background_tasks:
            self.run_peer_health_checks(context=context)

        ServerRegistry.set_server_for(self.id, self)
        if background_tasks:
            email_dispatcher = threading.Thread(
                target=self.email_notification_dispatcher, daemon=True
            )
            email_dispatcher.start()

    def email_notification_dispatcher(self) -> None:
        lock = threading.Lock()
        while True:
            # Use admin context to have access to the notifier obj
            context = AuthedServiceContext(
                server=self,
                credentials=self.verify_key,
                role=ServiceRole.ADMIN,
            )
            # Get notitifer settings
            notifier_settings = self.services.notifier.settings(
                context=context
            ).unwrap()
            lock.acquire()
            # Iterate over email_types and its queues
            # Ex: {'EmailRequest': {VerifyKey: [], VerifyKey: [], ...}}
            for email_template, email_queue in notifier_settings.email_queue.items():
                # Get the email frequency of that specific email type
                email_frequency = notifier_settings.email_frequency[email_template]
                for verify_key, queue in email_queue.items():
                    if self.services.notifier.is_time_to_dispatch(
                        email_frequency, datetime.now(timezone.utc)
                    ):
                        notifier_settings.send_batched_notification(
                            context=context, notification_queue=queue
                        ).unwrap()
                        notifier_settings.email_queue[email_template][verify_key] = []
                        self.services.notifier.stash.update(
                            credentials=self.verify_key, obj=notifier_settings
                        ).unwrap()
            lock.release()
            sleep(15)

    def set_log_level(self, log_level: int | str | None) -> None:
        def determine_log_level(
            log_level: str | int | None, default: int
        ) -> int | None:
            if log_level is None:
                return default
            if isinstance(log_level, str):
                level = logging.getLevelName(log_level.upper())
                if isinstance(level, str) and level.startswith("Level "):
                    level = logging.INFO  # defaults to info otherwise
                return level  # type: ignore
            return log_level

        default = logging.CRITICAL
        if self.deployment_type == DeploymentType.PYTHON:
            default = logging.CRITICAL
        elif self.dev_mode:  # if real deployment and dev mode
            default = logging.INFO

        self.log_level = determine_log_level(log_level, default)

        logging.getLogger().setLevel(self.log_level)

        if log_level == logging.DEBUG:
            # only do this if specifically set, very noisy
            logging.getLogger("uvicorn").setLevel(logging.DEBUG)
            logging.getLogger("uvicorn.access").setLevel(logging.DEBUG)
        else:
            logging.getLogger("uvicorn").setLevel(logging.CRITICAL)
            logging.getLogger("uvicorn.access").setLevel(logging.CRITICAL)

    @property
    def runs_in_docker(self) -> bool:
        path = "/proc/self/cgroup"
        return (
            os.path.exists("/.dockerenv")
            or os.path.isfile(path)
            and any("docker" in line for line in open(path))
        )

    def init_blob_storage(self, config: BlobStorageConfig | None = None) -> None:
        if config is None:
            client_config = OnDiskBlobStorageClientConfig(
                base_directory=self.get_temp_dir("blob")
            )
            config_ = OnDiskBlobStorageConfig(
                client_config=client_config,
                min_blob_size=os.getenv("MIN_SIZE_BLOB_STORAGE_MB", 1),
            )
        else:
            config_ = config
        self.blob_store_config = config_
        self.blob_storage_client = config_.client_type(config=config_.client_config)

        # relative
        from ..store.blob_storage.seaweedfs import SeaweedFSConfig

        if isinstance(config, SeaweedFSConfig) and self.signing_key:
            remote_profiles = self.services.blob_storage.remote_profile_stash.get_all(
                credentials=self.signing_key.verify_key, has_permission=True
            ).unwrap()
            for remote_profile in remote_profiles:
                self.blob_store_config.client_config.remote_profiles[
                    remote_profile.profile_name
                ] = remote_profile

        if self.dev_mode:
            if isinstance(self.blob_store_config, OnDiskBlobStorageConfig):
                logger.debug(
                    f"Using on-disk blob storage with path: "
                    f"{self.blob_store_config.client_config.base_directory}",
                )
            logger.debug(
                f"Minimum object size to be saved to the blob storage: "
                f"{self.blob_store_config.min_blob_size} (MB)."
            )

    def run_peer_health_checks(self, context: AuthedServiceContext) -> None:
        self.peer_health_manager = PeerHealthCheckTask()
        self.peer_health_manager.run(context=context)

    def stop(self) -> None:
        if self.peer_health_manager is not None:
            self.peer_health_manager.stop()

        for consumer_list in self.queue_manager.consumers.values():
            for c in consumer_list:
                c.close()
        for p in self.queue_manager.producers.values():
            p.close()

        self.queue_manager.producers.clear()
        self.queue_manager.consumers.clear()

        ServerRegistry.remove_server(self.id)

    def close(self) -> None:
        self.stop()

    def cleanup(self) -> None:
        self.stop()
        self.remove_temp_dir()

    def create_queue_config(
        self,
        n_consumers: int,
        create_producer: bool,
        consumer_type: ConsumerType,
        queue_port: int | None,
        queue_config: QueueConfig | None,
    ) -> QueueConfig:
        if queue_config:
            queue_config_ = queue_config
        elif queue_port is not None or n_consumers > 0 or create_producer:
            if not create_producer and queue_port is None:
                logger.warn("No queue port defined to bind consumers.")

            queue_config_ = ZMQQueueConfig(
                client_config=ZMQClientConfig(
                    create_producer=create_producer,
                    queue_port=queue_port,
                    n_consumers=n_consumers,
                ),
                consumer_type=consumer_type,
            )
        else:
            queue_config_ = ZMQQueueConfig()

        return queue_config_

    def init_queue_manager(self, queue_config: QueueConfig) -> None:
        MessageHandlers = [APICallMessageHandler]
        if self.is_subprocess:
            return None

        self.queue_manager = QueueManager(config=queue_config)
        for message_handler in MessageHandlers:
            queue_name = message_handler.queue_name
            # client config
            if getattr(queue_config.client_config, "create_producer", True):
                context = AuthedServiceContext(
                    server=self,
                    credentials=self.verify_key,
                    role=ServiceRole.ADMIN,
                )
                producer: QueueProducer = self.queue_manager.create_producer(
                    queue_name=queue_name,
                    queue_stash=self.queue_stash,
                    context=context,
                    worker_stash=self.worker_stash,
                )
                producer.run()

                address = producer.address
            else:
                port = queue_config.client_config.queue_port
                if port is not None:
                    address = get_queue_address(port)
                else:
                    address = None

            if address is None and queue_config.client_config.n_consumers > 0:
                raise ValueError("address unknown for consumers")

            service_name = queue_config.client_config.consumer_service

            if not service_name:
                # Create consumers for default worker pool
                create_default_worker_pool(self)
            else:
                # Create consumer for given worker pool
                syft_worker_uid = get_syft_worker_uid()
                logger.info(
                    f"Running as consumer with uid={syft_worker_uid} service={service_name}"
                )

                if syft_worker_uid:
                    self.add_consumer_for_service(
                        service_name=service_name,
                        syft_worker_id=UID(syft_worker_uid),
                        address=address,
                        message_handler=message_handler,
                    )

            if self.in_memory_workers:
                self.start_in_memory_workers(
                    address=address, message_handler=message_handler
                )

    def start_in_memory_workers(
        self, address: str, message_handler: type[AbstractMessageHandler]
    ) -> None:
        """Starts in-memory workers for the server."""

        worker_pools = self.pool_stash.get_all(credentials=self.verify_key).unwrap()
        for worker_pool in worker_pools:  # type: ignore
            # Skip the default worker pool
            if worker_pool.name == DEFAULT_WORKER_POOL_NAME:
                continue

            # Create consumers for each worker pool
            for linked_worker in worker_pool.worker_list:
                self.add_consumer_for_service(
                    service_name=worker_pool.name,
                    syft_worker_id=linked_worker.object_uid,
                    address=address,
                    message_handler=message_handler,
                )

    def add_consumer_for_service(
        self,
        service_name: str,
        syft_worker_id: UID,
        address: str,
        message_handler: type[AbstractMessageHandler] = APICallMessageHandler,
    ) -> None:
        consumer: QueueConsumer = self.queue_manager.create_consumer(
            message_handler,
            address=address,
            service_name=service_name,
            worker_stash=self.worker_stash,  # type: ignore
            syft_worker_id=syft_worker_id,
        )
        consumer.run()

    def remove_consumer_with_id(self, syft_worker_id: UID) -> None:
        for consumers in self.queue_manager.consumers.values():
            # Grab the list of consumers for the given queue
            consumer_to_pop = None
            for consumer_idx, consumer in enumerate(consumers):
                if consumer.syft_worker_id == syft_worker_id:
                    consumer.close()
                    consumer_to_pop = consumer_idx
                    break
            if consumer_to_pop is not None:
                consumers.pop(consumer_to_pop)

    @classmethod
    def named(
        cls: type[Server],
        *,  # Trasterisk
        name: str,
        processes: int = 0,
        reset: bool = False,
        server_type: str | ServerType = ServerType.DATASITE,
        server_side_type: str | ServerSideType = ServerSideType.HIGH_SIDE,
        deployment_type: str | DeploymentType = "remote",
        enable_warnings: bool = False,
        n_consumers: int = 0,
        thread_workers: bool = False,
        create_producer: bool = False,
        queue_port: int | None = None,
        dev_mode: bool = False,
        migrate: bool = False,
        in_memory_workers: bool = True,
        association_request_auto_approval: bool = False,
        background_tasks: bool = False,
        consumer_type: ConsumerType | None = None,
        db_url: str | None = None,
        db_config: DBConfig | None = None,
        log_level: int | None = None,
    ) -> Server:
        uid = get_named_server_uid(name)
        name_hash = hashlib.sha256(name.encode("utf8")).digest()
        key = SyftSigningKey(signing_key=SigningKey(name_hash))
        blob_storage_config = None

        server_type = ServerType(server_type)
        server_side_type = ServerSideType(server_side_type)

        return cls(
            name=name,
            id=uid,
            signing_key=key,
            processes=processes,
            server_type=server_type,
            server_side_type=server_side_type,
            deployment_type=deployment_type,
            enable_warnings=enable_warnings,
            blob_storage_config=blob_storage_config,
            queue_port=queue_port,
            n_consumers=n_consumers,
            thread_workers=thread_workers,
            create_producer=create_producer,
            dev_mode=dev_mode,
            migrate=migrate,
            in_memory_workers=in_memory_workers,
            reset=reset,
            association_request_auto_approval=association_request_auto_approval,
            background_tasks=background_tasks,
            consumer_type=consumer_type,
            db_url=db_url,
            db_config=db_config,
            log_level=log_level,
        )

    def is_root(self, credentials: SyftVerifyKey) -> bool:
        return credentials == self.verify_key

    @property
    def root_client(self) -> SyftClient:
        # relative
        from ..client.client import PythonConnection

        connection = PythonConnection(server=self)
        client_type = connection.get_client_type().unwrap()
        root_client = client_type(connection=connection, credentials=self.signing_key)

        if root_client.api.refresh_api_callback is not None:
            root_client.api.refresh_api_callback()

        return root_client

    def _find_klasses_pending_for_migration(
        self, object_types: list[SyftObject]
    ) -> list[SyftObject]:
        context = AuthedServiceContext(
            server=self,
            credentials=self.verify_key,
            role=ServiceRole.ADMIN,
        )
        migration_state_service = self.services.migration

        klasses_to_be_migrated = []

        for object_type in object_types:
            canonical_name = object_type.__canonical_name__
            object_version = object_type.__version__

            try:
                migration_state = migration_state_service.get_state(
                    context, canonical_name
                ).unwrap()
                if migration_state.current_version != migration_state.latest_version:
                    klasses_to_be_migrated.append(object_type)
            except NotFoundException:
                migration_state_service.register_migration_state(
                    context,
                    current_version=object_version,
                    canonical_name=canonical_name,
                )

        return klasses_to_be_migrated

    def find_and_migrate_data(
        self, document_store_object_types: list[type[SyftObject]] | None = None
    ) -> None:
        context = AuthedServiceContext(
            server=self,
            credentials=self.verify_key,
            role=ServiceRole.ADMIN,
        )
        return self.services.migration.migrate_data(
            context, document_store_object_types
        )

    @property
    def guest_client(self) -> SyftClient:
        return self.get_guest_client()

    @property
    def current_protocol(self) -> str | int:
        data_protocol = get_data_protocol()
        return data_protocol.latest_version

    def get_guest_client(self, verbose: bool = True) -> SyftClient:
        # relative
        from ..client.client import PythonConnection

        connection = PythonConnection(server=self)
        if verbose and self.server_side_type:
            message: str = (
                f"Logged into <{self.name}: {self.server_side_type.value.capitalize()} "
            )
            if self.server_type:
                message += f"side {self.server_type.value.capitalize()} > as GUEST"
            logger.debug(message)

        client_type = connection.get_client_type().unwrap()

        guest_client = client_type(
            connection=connection, credentials=SyftSigningKey.generate()
        )
        if guest_client.api.refresh_api_callback is not None:
            guest_client.api.refresh_api_callback()
        return guest_client

    def __repr__(self) -> str:
        service_string = ""
        if not self.is_subprocess:
            services = [
                service.__class__.__name__ for service in self.initialized_services
            ]
            service_string = ", ".join(sorted(services))
            service_string = f"\n\nServices:\n{service_string}"
        return f"{type(self).__name__}: {self.name} - {self.id} - {self.server_type}{service_string}"

    def post_init(self) -> None:
        context = AuthedServiceContext(
            server=self, credentials=self.verify_key, role=ServiceRole.ADMIN
        )
        AuthServerContextRegistry.set_server_context(
            server_uid=self.id, user_verify_key=self.verify_key, context=context
        )

        if "usercodeservice" in self.service_path_map:
            self.services.user_code.load_user_code(context=context)

        def reload_user_code() -> None:
            self.services.user_code.load_user_code(context=context)

        ti = thread_ident()
        if ti is not None:
            CODE_RELOADER[ti] = reload_user_code

    def init_stores(self, db_config: DBConfig) -> DBManager:
        if isinstance(db_config, SQLiteDBConfig):
            db = SQLiteDBManager(
                config=db_config,
                server_uid=self.id,
                root_verify_key=self.verify_key,
            )
        elif isinstance(db_config, PostgresDBConfig):
            db = PostgresDBManager(  # type: ignore
                config=db_config,
                server_uid=self.id,
                root_verify_key=self.verify_key,
            )
        else:
            raise SyftException(public_message=f"Unsupported DB config: {db_config}")

        self.queue_stash = QueueStash(store=db)

        print(f"Using {db_config.__class__.__name__} and {db_config.connection_string}")

        return db

    @property
    def job_stash(self) -> JobStash:
        return self.services.job.stash

    @property
    def output_stash(self) -> OutputStash:
        return self.services.output.stash

    @property
    def worker_stash(self) -> WorkerStash:
        return self.services.worker.stash

    @property
    def service_path_map(self) -> dict[str, AbstractService]:
        return self.services.service_path_map

    @property
    def initialized_services(self) -> list[AbstractService]:
        return self.services.services

    def get_service_method(self, path_or_func: str | Callable) -> Callable:
        if callable(path_or_func):
            path_or_func = path_or_func.__qualname__
        return self._get_service_method_from_path(path_or_func)

    def get_service(self, path_or_func: str | Callable) -> AbstractService:
        return self.services.get_service(path_or_func)

    @as_result(ValueError)
    def get_stash(self, object_type: SyftT) -> ObjectStash[SyftT]:
        if object_type not in self.services.stashes:
            raise ValueError(f"Stash for {object_type} not found.")
        return self.services.stashes[object_type]

    def _get_service_method_from_path(self, path: str) -> Callable:
        path_list = path.split(".")
        method_name = path_list.pop()
        service_obj = self.services._get_service_from_path(path=path)

        return getattr(service_obj, method_name)

    def get_temp_dir(self, dir_name: str = "") -> Path:
        """
        Get a temporary directory unique to the server.
        Provide all dbs, blob dirs, and locks using this directory.
        """
        return get_temp_dir_for_server(self.id, dir_name)

    def remove_temp_dir(self) -> None:
        """
        Remove the temporary directory for this server.
        """
        remove_temp_dir_for_server(self.id)

    def update_self(self, settings: ServerSettings) -> None:
        updateable_attrs = (
            ServerSettingsUpdate.model_fields.keys()
            - PartialSyftObject.model_fields.keys()
        )
        for attr_name in updateable_attrs:
            attr = getattr(settings, attr_name)
            if attr is not Empty:
                setattr(self, attr_name, attr)

    # NOTE: Some workflows currently expect the settings to be available,
    # even though they might not be defined yet. Because of this, we need to check
    # if the settings table is already defined. This function is basically a copy
    # of the settings property but ignoring stash error in case settings doesn't exist yet.
    # it should be removed once the settings are refactored and the inconsistencies between
    # settings and services are resolved.
    def get_settings(self) -> ServerSettings | None:
        if self._settings:
            return self._settings  # type: ignore
        if self.signing_key is None:
            raise ValueError(f"{self} has no signing key")

        settings_stash = self.services.settings.stash

        try:
            settings = settings_stash.get_all(self.signing_key.verify_key).unwrap()

            if len(settings) > 0:
                setting = settings[0]
                self.update_self(setting)
                self._settings = setting
                return setting
            else:
                return None

        except SyftException:
            return None

    @property
    def settings(self) -> ServerSettings:
        if self.signing_key is None:
            raise ValueError(f"{self} has no signing key")

        settings_stash = self.services.settings.stash
        error_msg = f"Cannot get server settings for '{self.name}'"

        all_settings = settings_stash.get_all(self.signing_key.verify_key).unwrap(
            public_message=error_msg
        )

        if len(all_settings) == 0:
            raise SyftException(public_message=error_msg)

        settings = all_settings[0]
        self.update_self(settings)
        return settings

    @property
    def metadata(self) -> ServerMetadata:
        show_warnings = self.enable_warnings
        settings_data = self.settings
        name = settings_data.name
        organization = settings_data.organization
        description = settings_data.description
        show_warnings = settings_data.show_warnings
        server_type = (
            settings_data.server_type.value if settings_data.server_type else ""
        )
        server_side_type = (
            settings_data.server_side_type.value
            if settings_data.server_side_type
            else ""
        )
        eager_execution_enabled = settings_data.eager_execution_enabled

        return ServerMetadata(
            name=name,
            id=self.id,
            verify_key=self.verify_key,
            highest_version=SYFT_OBJECT_VERSION_1,
            lowest_version=SYFT_OBJECT_VERSION_1,
            syft_version=__version__,
            description=description,
            organization=organization,
            server_type=server_type,
            server_side_type=server_side_type,
            show_warnings=show_warnings,
            eager_execution_enabled=eager_execution_enabled,
            min_size_blob_storage_mb=self.blob_store_config.min_blob_size,
        )

    @property
    def icon(self) -> str:
        return "🦾"

    @property
    def verify_key(self) -> SyftVerifyKey:
        if self.signing_key is None:
            raise ValueError(f"{self} has no signing key")
        return self.signing_key.verify_key

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return False

        if self.id != other.id:
            return False

        return True

    def await_future(self, credentials: SyftVerifyKey, uid: UID) -> QueueItem:
        # stdlib

        # relative
        from ..service.queue.queue import Status

        while True:
            result = self.queue_stash.pop_on_complete(credentials, uid).unwrap()
            if result.status == Status.COMPLETED:
                return result
            sleep(0.1)

    @instrument
    def resolve_future(self, credentials: SyftVerifyKey, uid: UID) -> QueueItem:
        queue_obj = self.queue_stash.pop_on_complete(credentials, uid).unwrap()
        queue_obj._set_obj_location_(
            server_uid=self.id,
            credentials=credentials,
        )
        return queue_obj

    @instrument
    def forward_message(
        self, api_call: SyftAPICall | SignedSyftAPICall
    ) -> Result | QueueItem | SyftObject | Any:
        server_uid = api_call.message.server_uid
        if "networkservice" not in self.service_path_map:
            raise SyftException(
                public_message=(
                    "Server has no network service so we can't "
                    f"forward this message to {server_uid}"
                )
            )

        client = None
        peer = self.services.network.stash.get_by_uid(
            self.verify_key, server_uid
        ).unwrap()

        # Since we have several routes to a peer
        # we need to cache the client for a given server_uid along with the route
        peer_cache_key = hash(server_uid) + hash(peer.pick_highest_priority_route())
        if peer_cache_key in self.peer_client_cache:
            client = self.peer_client_cache[peer_cache_key]
        else:
            context = AuthedServiceContext(
                server=self, credentials=api_call.credentials
            )

            client = peer.client_with_context(context=context).unwrap(
                public_message=f"Failed to create remote client for peer: {peer.id}"
            )
            self.peer_client_cache[peer_cache_key] = client

        if client:
            message: SyftAPICall = api_call.message
            if message.path == "metadata":
                result = client.metadata
            elif message.path == "login":
                result = client.connection.login(**message.kwargs)
            elif message.path == "register":
                result = client.connection.register(**message.kwargs)
            elif message.path == "api":
                result = client.connection.get_api(**message.kwargs)
            else:
                signed_result = client.connection.make_call(api_call)
                result = debox_signed_syftapicall_response(
                    signed_result=signed_result
                ).unwrap()

                # relative
                from ..store.blob_storage import BlobRetrievalByURL

                if isinstance(result, BlobRetrievalByURL | SeaweedFSBlobDeposit):
                    result.proxy_server_uid = peer.id

            return result

        raise SyftException(public_message=(f"Server has no route to {server_uid}"))

    def get_role_for_credentials(self, credentials: SyftVerifyKey) -> ServiceRole:
        return self.services.user.get_role_for_credentials(
            credentials=credentials
        ).unwrap()

    @instrument
    def handle_api_call(
        self,
        api_call: SyftAPICall | SignedSyftAPICall,
        job_id: UID | None = None,
        check_call_location: bool = True,
    ) -> SignedSyftAPICall:
        # Get the result
        result = self.handle_api_call_with_unsigned_result(
            api_call, job_id=job_id, check_call_location=check_call_location
        )
        # Sign the result
        signed_result = SyftAPIData(data=result).sign(self.signing_key)

        return signed_result

    def handle_api_call_with_unsigned_result(
        self,
        api_call: SyftAPICall | SignedSyftAPICall,
        job_id: UID | None = None,
        check_call_location: bool = True,
    ) -> Result | QueueItem | SyftObject | SyftError:
        if self.required_signed_calls and isinstance(api_call, SyftAPICall):
            raise SyftException(
                public_message=f"You sent a {type(api_call)}. This server requires SignedSyftAPICall."
            )
        else:
            if not api_call.is_valid:
                raise SyftException(public_message="Your message signature is invalid")

        if api_call.message.server_uid != self.id and check_call_location:
            return self.forward_message(api_call=api_call)

        if api_call.message.path == "queue":
            return self.resolve_future(
                credentials=api_call.credentials, uid=api_call.message.kwargs["uid"]
            )

        if api_call.message.path == "metadata":
            return self.metadata

        result = None
        is_blocking = api_call.message.blocking

        credentials: SyftVerifyKey = api_call.credentials
        role = self.get_role_for_credentials(credentials=credentials)
        context = AuthedServiceContext(
            server=self,
            credentials=credentials,
            role=role,
            job_id=job_id,
            is_blocking_api_call=is_blocking,
        )

        if is_blocking or self.is_subprocess:
            api_call = api_call.message

            role = self.get_role_for_credentials(credentials=credentials)
            settings = self.get_settings()
            # TODO: This instance check should be removed once we can ensure that
            # self.settings will always return a ServerSettings object.
            if (
                settings is not None
                and isinstance(settings, ServerSettings)
                and not settings.allow_guest_sessions
                and role == ServiceRole.GUEST
            ):
                raise SyftException(
                    public_message="Server doesn't allow guest sessions."
                )
            context = AuthedServiceContext(
                server=self,
                credentials=credentials,
                role=role,
                job_id=job_id,
                is_blocking_api_call=is_blocking,
            )

            AuthServerContextRegistry.set_server_context(self.id, context, credentials)

            user_config_registry = UserServiceConfigRegistry.from_role(role)

            if api_call.path not in user_config_registry:
                if ServiceConfigRegistry.path_exists(api_call.path):
                    raise SyftException(
                        public_message=f"As a `{role}`, "
                        f"you have no access to: {api_call.path}"
                    )
                else:
                    raise SyftException(
                        public_message=f"API call not in registered services: {api_call.path}"
                    )

            _private_api_path = user_config_registry.private_path_for(api_call.path)
            method = self.get_service_method(_private_api_path)
            try:
                logger.info(f"API Call: {api_call}")

                result = method(context, *api_call.args, **api_call.kwargs)

                if isinstance(result, SyftError):
                    raise TypeError(
                        "Don't return a SyftError, raise SyftException instead"
                    )
                if not isinstance(result, SyftSuccess):
                    result = SyftSuccess(message="", value=result)
                result.add_warnings_from_context(context)
                tb = None
            except Exception as e:
                include_traceback = (
                    self.dev_mode or role.value >= ServiceRole.DATA_OWNER.value
                )
                result = SyftError.from_exception(
                    context=context, exc=e, include_traceback=include_traceback
                )
                if not include_traceback:
                    # then at least log it server side
                    if isinstance(e, SyftException):
                        tb = e.get_tb(context, overwrite_permission=True)
                    else:
                        tb = traceback.format_exc()
                    logger.debug(
                        f"Exception (hidden from DS) happened on the server side:\n{tb}"
                    )
        else:
            try:
                return self.add_api_call_to_queue(api_call)
            except SyftException as e:
                return SyftError.from_exception(context=context, exc=e)
            except Exception:
                result = SyftError(
                    message=f"Exception calling {api_call.path}. {traceback.format_exc()}"
                )
                tb = traceback.format_exc()
            if (
                isinstance(result, SyftError)
                and role.value < ServiceRole.DATA_OWNER.value
            ):
                print(f"Exception (hidden from DS) happened on the server side:\n{tb}")
        return result

    def add_api_endpoint_execution_to_queue(
        self,
        credentials: SyftVerifyKey,
        method: str,
        path: str,
        log_id: UID,
        *args: Any,
        worker_pool_name: str | None = None,
        **kwargs: Any,
    ) -> Job:
        job_id = UID()
        task_uid = UID()
        worker_settings = WorkerSettings.from_server(server=self)

        if worker_pool_name is None:
            worker_pool_name = self.get_default_worker_pool().unwrap()
        else:
            worker_pool_name = self.get_worker_pool_by_name(worker_pool_name).unwrap()

        # Create a Worker pool reference object
        worker_pool_ref = LinkedObject.from_obj(
            worker_pool_name,
            service_type=SyftWorkerPoolService,
            server_uid=self.id,
        )
        queue_item = APIEndpointQueueItem(
            id=task_uid,
            method=method,
            server_uid=self.id,
            syft_client_verify_key=credentials,
            syft_server_location=self.id,
            job_id=job_id,
            worker_settings=worker_settings,
            args=args,
            kwargs={"path": path, "log_id": log_id, **kwargs},
            has_execute_permissions=True,
            worker_pool=worker_pool_ref,  # set worker pool reference as part of queue item
        )

        action = Action.from_api_endpoint_execution()
        return self.add_queueitem_to_queue(
            queue_item=queue_item,
            credentials=credentials,
            action=action,
            job_type=JobType.TWINAPIJOB,
        ).unwrap()

    def get_worker_pool_ref_by_name(
        self, credentials: SyftVerifyKey, worker_pool_name: str | None = None
    ) -> LinkedObject:
        # If worker pool id is not set, then use default worker pool
        # Else, get the worker pool for given uid
        if worker_pool_name is None:
            worker_pool = self.get_default_worker_pool().unwrap()
        else:
            worker_pool = self.pool_stash.get_by_name(
                credentials, worker_pool_name
            ).unwrap()

        # Create a Worker pool reference object
        worker_pool_ref = LinkedObject.from_obj(
            worker_pool,
            service_type=SyftWorkerPoolService,
            server_uid=self.id,
        )
        return worker_pool_ref

    @instrument
    @as_result(SyftException)
    def add_action_to_queue(
        self,
        action: Action,
        credentials: SyftVerifyKey,
        parent_job_id: UID | None = None,
        has_execute_permissions: bool = False,
        worker_pool_name: str | None = None,
    ) -> Job:
        job_id = UID()
        task_uid = UID()
        worker_settings = WorkerSettings.from_server(server=self)

        # Extract worker pool id from user code
        if action.user_code_id is not None:
            user_code = self.user_code_stash.get_by_uid(
                credentials=credentials, uid=action.user_code_id
            ).unwrap()
            if user_code is not None:
                worker_pool_name = user_code.worker_pool_name

        worker_pool_ref = self.get_worker_pool_ref_by_name(
            credentials, worker_pool_name
        )
        queue_item = ActionQueueItem(
            id=task_uid,
            server_uid=self.id,
            syft_client_verify_key=credentials,
            syft_server_location=self.id,
            job_id=job_id,
            worker_settings=worker_settings,
            args=[],
            kwargs={"action": action},
            has_execute_permissions=has_execute_permissions,
            worker_pool=worker_pool_ref,  # set worker pool reference as part of queue item
        )
        user_id = self.services.user.get_user_id_for_credentials(credentials).unwrap()

        return self.add_queueitem_to_queue(
            queue_item=queue_item,
            credentials=credentials,
            action=action,
            parent_job_id=parent_job_id,
            user_id=user_id,
        ).unwrap()

    @instrument
    @as_result(SyftException)
    def add_queueitem_to_queue(
        self,
        *,
        queue_item: QueueItem,
        credentials: SyftVerifyKey,
        action: Action | None = None,
        parent_job_id: UID | None = None,
        user_id: UID | None = None,
        log_id: UID | None = None,
        job_type: JobType = JobType.JOB,
    ) -> Job:
        if log_id is None:
            log_id = UID()
        role = self.get_role_for_credentials(credentials=credentials)
        context = AuthedServiceContext(server=self, credentials=credentials, role=role)

        result_obj = ActionObject.empty()
        if action is not None:
            result_obj = ActionObject.obj_not_ready(
                id=action.result_id,
                syft_server_location=self.id,
                syft_client_verify_key=credentials,
            )
            result_obj.id = action.result_id
            result_obj.syft_resolved = False
            result_obj.syft_server_location = self.id
            result_obj.syft_client_verify_key = credentials

            if not self.services.action.stash.exists(
                credentials=credentials, uid=action.result_id
            ):
                self.services.action.set_result_to_store(
                    result_action_object=result_obj,
                    context=context,
                ).unwrap()

        job = Job(
            id=queue_item.job_id,
            result=result_obj,
            server_uid=self.id,
            syft_client_verify_key=credentials,
            syft_server_location=self.id,
            log_id=log_id,
            parent_job_id=parent_job_id,
            action=action,
            requested_by=user_id,
            job_type=job_type,
            endpoint=queue_item.kwargs.get("path", None),
        )

        # 🟡 TODO 36: Needs distributed lock
        self.job_stash.set(credentials, job).unwrap()
        self.queue_stash.set_placeholder(credentials, queue_item).unwrap()

        self.services.log.add(context, log_id, queue_item.job_id)

        return job

    def _sort_jobs(self, jobs: list[Job]) -> list[Job]:
        job_datetimes = {}
        for job in jobs:
            try:
                d = datetime.strptime(job.creation_time, DATETIME_FORMAT)
            except Exception:
                d = datetime(MINYEAR, 1, 1)
            job_datetimes[job.id] = d

        jobs.sort(
            key=lambda job: (job.status != JobStatus.COMPLETED, job_datetimes[job.id]),
            reverse=True,
        )

        return jobs

    @as_result(SyftException)
    def _get_existing_user_code_jobs(
        self, context: AuthedServiceContext, user_code_id: UID
    ) -> list[Job]:
        jobs = self.services.job.get_by_user_code_id(
            context=context, user_code_id=user_code_id
        )
        return self._sort_jobs(jobs)

    def _is_usercode_call_on_owned_kwargs(
        self,
        context: AuthedServiceContext,
        api_call: SyftAPICall,
        user_code_id: UID,
    ) -> bool:
        if api_call.path != "code.call":
            return False
        return self.services.user_code.is_execution_on_owned_args(
            context, user_code_id, api_call.kwargs
        )

    @instrument
    def add_api_call_to_queue(
        self, api_call: SyftAPICall, parent_job_id: UID | None = None
    ) -> SyftSuccess:
        unsigned_call = api_call
        if isinstance(api_call, SignedSyftAPICall):
            unsigned_call = api_call.message

        credentials = api_call.credentials
        context = AuthedServiceContext(
            server=self,
            credentials=credentials,
            role=self.get_role_for_credentials(credentials=credentials),
        )

        is_user_code = unsigned_call.path == "code.call"

        service_str, method_str = unsigned_call.path.split(".")

        action = None
        if is_user_code:
            action = Action.from_api_call(unsigned_call)
            user_code_id = action.user_code_id

            user = self.services.user.get_current_user(context)
            user = cast(UserView, user)

            is_execution_on_owned_kwargs_allowed = (
                user.mock_execution_permission or context.role == ServiceRole.ADMIN
            )
            is_usercode_call_on_owned_kwargs = self._is_usercode_call_on_owned_kwargs(
                context, unsigned_call, user_code_id
            )
            # Low side does not execute jobs, unless this is a mock execution
            if (
                not is_usercode_call_on_owned_kwargs
                and self.server_side_type == ServerSideType.LOW_SIDE
            ):
                try:
                    existing_jobs = self._get_existing_user_code_jobs(
                        context, user_code_id
                    ).unwrap()

                    if len(existing_jobs) > 0:
                        # relative
                        from ..util.util import prompt_warning_message

                        prompt_warning_message(
                            "There are existing jobs for this user code, returning the latest one"
                        )
                        return SyftSuccess(
                            message="Found multiple existing jobs, got last",
                            value=existing_jobs[-1],
                        )
                    else:
                        raise SyftException(
                            public_message="Please wait for the admin to allow the execution of this code"
                        )
                except Exception as e:
                    raise SyftException.from_exception(e)
            elif (
                is_usercode_call_on_owned_kwargs
                and not is_execution_on_owned_kwargs_allowed
            ):
                raise SyftException(
                    public_message="You do not have the permissions for mock execution, please contact the admin"
                )

            job = self.add_action_to_queue(
                action, api_call.credentials, parent_job_id=parent_job_id
            ).unwrap()

            return SyftSuccess(message="Succesfully queued job", value=job)

        else:
            worker_settings = WorkerSettings.from_server(server=self)
            worker_pool_ref = self.get_worker_pool_ref_by_name(credentials=credentials)
            queue_item = QueueItem(
                id=UID(),
                server_uid=self.id,
                syft_client_verify_key=api_call.credentials,
                syft_server_location=self.id,
                job_id=UID(),
                worker_settings=worker_settings,
                service=service_str,
                method=method_str,
                args=unsigned_call.args,
                kwargs=unsigned_call.kwargs,
                worker_pool=worker_pool_ref,
            )
            return self.add_queueitem_to_queue(
                queue_item=queue_item,
                credentials=api_call.credentials,
                action=None,
                parent_job_id=parent_job_id,
            ).unwrap()

    @property
    def pool_stash(self) -> SyftWorkerPoolStash:
        return self.services.syft_worker_pool.stash

    @property
    def user_code_stash(self) -> UserCodeStash:
        return self.services.user_code.stash

    @as_result(NotFoundException)
    def get_default_worker_pool(self) -> WorkerPool | None:
        return self.pool_stash.get_by_name(
            credentials=self.verify_key,
            pool_name=self.settings.default_worker_pool,
        ).unwrap()

    @as_result(NotFoundException)
    def get_worker_pool_by_name(self, name: str) -> WorkerPool:
        return self.pool_stash.get_by_name(
            credentials=self.verify_key, pool_name=name
        ).unwrap()

    @instrument
    def get_api(
        self,
        for_user: SyftVerifyKey | None = None,
        communication_protocol: PROTOCOL_TYPE | None = None,
    ) -> SyftAPI:
        return SyftAPI.for_user(
            server=self,
            user_verify_key=for_user,
            communication_protocol=communication_protocol,
        )

    def get_method_with_context(
        self, function: Callable, context: ServerServiceContext
    ) -> Callable:
        method = self.get_service_method(function)
        return partial(method, context=context)

    def get_unauthed_context(
        self, login_credentials: UserLoginCredentials
    ) -> ServerServiceContext:
        return UnauthedServiceContext(server=self, login_credentials=login_credentials)

    @as_result(SyftException, StashException)
    def create_initial_settings(self, admin_email: str) -> ServerSettings:
        settings_stash = self.services.settings.stash

        if self.signing_key is None:
            logger.debug("create_initial_settings failed as there is no signing key")
            raise SyftException(
                public_message="create_initial_settings failed as there is no signing key"
            )

        settings_exists = settings_stash.get_all(self.signing_key.verify_key).unwrap()

        if settings_exists:
            server_settings = settings_exists[0]
            if server_settings.__version__ != ServerSettings.__version__:
                context = Context()
                server_settings = server_settings.migrate_to(
                    ServerSettings.__version__, context
                )
                settings_stash.delete_by_uid(
                    self.signing_key.verify_key, server_settings.id
                ).unwrap()
                settings_stash.set(
                    self.signing_key.verify_key, server_settings
                ).unwrap()
            self.name = server_settings.name
            self.association_request_auto_approval = (
                server_settings.association_request_auto_approval
            )
            return server_settings
        else:
            # Currently we allow automatic user registration on enclaves,
            # as enclaves do not have superusers
            if self.server_type == ServerType.ENCLAVE:
                flags.CAN_REGISTER = True

            new_settings = ServerSettings(
                id=self.id,
                name=self.name,
                verify_key=self.verify_key,
                server_type=self.server_type,
                deployed_on=datetime.now().date().strftime("%m/%d/%Y"),
                signup_enabled=flags.CAN_REGISTER,
                admin_email=admin_email,
                server_side_type=self.server_side_type.value,  # type: ignore
                show_warnings=self.enable_warnings,
                association_request_auto_approval=self.association_request_auto_approval,
                default_worker_pool=get_default_worker_pool_name(),
                notifications_enabled=False,
            )

            return settings_stash.set(
                credentials=self.signing_key.verify_key, obj=new_settings
            ).unwrap()


class ServerRegistry:
    __server_registry__: dict[UID, Server] = {}

    @classmethod
    def set_server_for(
        cls,
        server_uid: UID | str,
        server: Server,
    ) -> None:
        if isinstance(server_uid, str):
            server_uid = UID.from_string(server_uid)

        cls.__server_registry__[server_uid] = server

    @classmethod
    def server_for(cls, server_uid: UID) -> Server:
        return cls.__server_registry__.get(server_uid, None)

    @classmethod
    def get_all_servers(cls) -> list[Server]:
        return list(cls.__server_registry__.values())

    @classmethod
    def remove_server(cls, server_uid: UID) -> None:
        if server_uid in cls.__server_registry__:
            del cls.__server_registry__[server_uid]


def get_default_worker_tag_by_env(dev_mode: bool = False) -> str | None:
    if in_kubernetes():
        return get_default_worker_image()
    elif dev_mode:
        return "local-dev"
    else:
        return __version__


def create_default_worker_pool(server: Server) -> None:
    credentials = server.verify_key
    pull_image = not server.dev_mode
    image_stash = server.services.syft_worker_image.stash
    default_pool_name = server.settings.default_worker_pool

    try:
        default_worker_pool = server.get_default_worker_pool().unwrap(
            public_message="Failed to get default worker pool"
        )
    except SyftException:
        default_worker_pool = None

    default_worker_tag = get_default_worker_tag_by_env(server.dev_mode)
    default_worker_pool_pod_annotations = get_default_worker_pool_pod_annotations()
    default_worker_pool_pod_labels = get_default_worker_pool_pod_labels()
    worker_count = get_default_worker_pool_count(server)
    context = AuthedServiceContext(
        server=server,
        credentials=credentials,
        role=ServiceRole.ADMIN,
    )

    logger.info(f"Creating default worker image with tag='{default_worker_tag}'. ")
    # Get/Create a default worker SyftWorkerImage
    # TODO: MERGE: Unwrap without public message?
    default_image = create_default_image(
        credentials=credentials,
        image_stash=image_stash,
        tag=default_worker_tag,
        in_kubernetes=in_kubernetes(),
    ).unwrap(public_message="Failed to create default worker image")

    if not default_image.is_built:
        logger.info(f"Building default worker image with tag={default_worker_tag}. ")
        # Build the Image for given tag
        result = server.services.worker_image.build(
            context,
            image_uid=default_image.id,
            tag=DEFAULT_WORKER_IMAGE_TAG,
            pull_image=pull_image,
        )

    # Create worker pool if it doesn't exists
    logger.info(
        "Setting up worker pool"
        f"name={default_pool_name} "
        f"workers={worker_count} "
        f"image_uid={default_image.id} "
        f"in_memory={server.in_memory_workers}. "
    )
    if default_worker_pool is None:
        worker_to_add_ = worker_count
        result = server.services.syft_worker_pool.launch(
            context,
            pool_name=default_pool_name,
            image_uid=default_image.id,
            num_workers=worker_count,
            pod_annotations=default_worker_pool_pod_annotations,
            pod_labels=default_worker_pool_pod_labels,
        )
    else:
        # Else add a worker to existing worker pool
        worker_to_add_ = max(default_worker_pool.max_count, worker_count) - len(
            default_worker_pool.worker_list
        )
        if worker_to_add_ > 0:
            result = server.services.syft_worker_pool.add_workers(
                context=context,
                number=worker_to_add_,
                pool_name=default_pool_name,
            )
        else:
            return None

    for n in range(worker_to_add_):
        container_status = result[n]
        if container_status.error:
            logger.error(
                f"Failed to create container: Worker: {container_status.worker},"
                f"Error: {container_status.error}"
            )
            return None

    logger.info("Created default worker pool.")
    return None



# stdlib
import logging

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.db import DBManager
from ...store.linked_obj import LinkedObject
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.uid import UID
from ..context import AuthedServiceContext
from ..notification.email_templates import EmailTemplate
from ..notification.email_templates import RequestEmailTemplate
from ..notification.email_templates import RequestUpdateEmailTemplate
from ..notification.notification_service import CreateNotification
from ..notifier.notifier_enums import NOTIFIERS
from ..notifier.notifier_service import RateLimitException
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from .request import Change
from .request import Request
from .request import RequestInfo
from .request import RequestInfoFilter
from .request import RequestStatus
from .request import SubmitRequest
from .request_stash import RequestStash

logger = logging.getLogger(__name__)


@serializable(canonical_name="RequestService", version=1)
class RequestService(AbstractService):
    stash: RequestStash

    def __init__(self, store: DBManager) -> None:
        self.stash = RequestStash(store=store)

    @service_method(path="request.submit", name="submit", roles=GUEST_ROLE_LEVEL)
    def submit(
        self,
        context: AuthedServiceContext,
        request: SubmitRequest,
        send_message: bool = True,
        reason: str | None = "",
    ) -> Request:
        """Submit a Request"""
        request = request.to(Request, context=context)
        request = self.stash.set(
            context.credentials,
            request,
        ).unwrap()

        root_verify_key = context.server.services.user.root_verify_key

        if send_message:
            message_subject = f"Result to request {str(request.id)[:4]}...{str(request.id)[-3:]}\
                has been successfully deposited."
            self._send_email_notification(
                context=context,
                message_subject=message_subject if not reason else reason,
                request=request,
                to_user_verify_key=root_verify_key,
                email_template=RequestEmailTemplate,
            )
        return request

    @service_method(
        path="request.get_by_uid", name="get_by_uid", roles=DATA_SCIENTIST_ROLE_LEVEL
    )
    def get_by_uid(self, context: AuthedServiceContext, uid: UID) -> Request:
        return self.stash.get_by_uid(context.credentials, uid).unwrap()

    @service_method(
        path="request.get_all", name="get_all", roles=DATA_SCIENTIST_ROLE_LEVEL
    )
    def get_all(self, context: AuthedServiceContext) -> list[Request]:
        requests = self.stash.get_all(context.credentials).unwrap()
        # TODO remove once sorting is handled by the stash
        requests.sort(key=lambda x: (x.request_time, x.id), reverse=True)

        return requests

    # DIRTY METHOD: DELETE AFTER DATABASE UPGRADE
    @service_method(
        path="request.get_all_approved",
        name="get_all_approved",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_all_approved(self, context: AuthedServiceContext) -> list[Request]:
        requests = self.stash.get_all(context.credentials).unwrap()
        # TODO remove once sorting is handled by the stash
        requests = [
            request
            for request in requests
            if request.get_status(context) == RequestStatus.APPROVED
        ]
        requests.sort(key=lambda x: (x.request_time, x.id), reverse=True)

        return requests

    # DIRTY METHOD: DELETE AFTER DATABASE UPGRADE
    @service_method(
        path="request.get_all_rejected",
        name="get_all_rejected",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_all_rejected(self, context: AuthedServiceContext) -> list[Request]:
        requests = self.stash.get_all(context.credentials).unwrap()
        # TODO remove once sorting is handled by the stash
        requests = [
            request
            for request in requests
            if request.get_status(context) == RequestStatus.REJECTED
        ]
        requests.sort(key=lambda x: (x.request_time, x.id), reverse=True)

        return requests

    # DIRTY METHOD: DELETE AFTER DATABASE UPGRADE
    @service_method(
        path="request.get_all_pending",
        name="get_all_pending",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_all_pending(self, context: AuthedServiceContext) -> list[Request]:
        requests = self.stash.get_all(context.credentials).unwrap()
        # TODO remove once sorting is handled by the stash
        requests = [
            request
            for request in requests
            if request.get_status(context) == RequestStatus.PENDING
        ]
        requests.sort(key=lambda x: (x.request_time, x.id), reverse=True)

        return requests

    @service_method(path="request.get_all_info", name="get_all_info")
    def get_all_info(
        self,
        context: AuthedServiceContext,
        page_index: int | None = 0,
        page_size: int | None = 0,
    ) -> list[list[RequestInfo]] | list[RequestInfo]:
        """Get the information of all requests"""
        result = self.stash.get_all(context.credentials).unwrap()
        requests: list[RequestInfo] = []
        for req in result:
            user = context.server.services.user.get_by_verify_key(
                req.requesting_user_verify_key
            ).unwrap()
            message = context.server.services.notification.filter_by_obj(
                context=context, obj_uid=req.id
            ).unwrap()
            requests.append(RequestInfo(user=user, request=req, notification=message))
        if not page_size:
            return requests

        # If chunk size is defined, then split list into evenly sized chunks
        chunked_requests: list[list[RequestInfo]] = [
            requests[i : i + page_size] for i in range(0, len(requests), page_size)
        ]
        if page_index:
            return chunked_requests[page_index]
        else:
            return chunked_requests

    @service_method(path="request.add_changes", name="add_changes")
    def add_changes(
        self, context: AuthedServiceContext, uid: UID, changes: list[Change]
    ) -> Request:
        request = self.stash.get_by_uid(
            credentials=context.credentials, uid=uid
        ).unwrap()
        request.changes.extend(changes)
        return self.save(context=context, request=request)

    @service_method(path="request.filter_all_info", name="filter_all_info")
    def filter_all_info(
        self,
        context: AuthedServiceContext,
        request_filter: RequestInfoFilter,
        page_index: int | None = 0,
        page_size: int | None = 0,
    ) -> list[RequestInfo]:
        """Filter Request"""
        result = self.get_all_info(context)

        requests = list(
            filter(lambda res: (request_filter.name in res.user.name), result)
        )

        # TODO: Move chunking to a function?
        # If chunk size is defined, then split list into evenly sized chunks
        if page_size:
            requests = [
                requests[i : i + page_size] for i in range(0, len(requests), page_size)
            ]
            if page_index is not None:
                # Return the proper slice using chunk_index
                requests = requests[page_index]

        return requests

    @service_method(path="request.apply", name="apply", unwrap_on_success=False)
    def apply(
        self,
        context: AuthedServiceContext,
        uid: UID,
        **kwargs: dict,
    ) -> SyftSuccess:
        request: Request = self.stash.get_by_uid(context.credentials, uid).unwrap()

        context.extra_kwargs = kwargs
        result = request.apply(context=context).unwrap()
        request_notification = context.server.services.notification.filter_by_obj(
            context=context, obj_uid=uid
        ).unwrap()

        if not request.get_status(context) == RequestStatus.PENDING:
            if request_notification is not None:
                context.server.services.notification.mark_as_read(
                    context=context, uid=request_notification.id
                )

                self._send_email_notification(
                    context=context,
                    message_subject=f"Your request ({str(uid)[:4]}) has been approved. ",
                    request=request,
                    to_user_verify_key=request.requesting_user_verify_key,
                    email_template=RequestUpdateEmailTemplate,
                )
        return result

    @as_result(SyftException, RateLimitException)
    def _send_email_notification(
        self,
        *,
        context: AuthedServiceContext,
        request: Request,
        message_subject: str,
        to_user_verify_key: SyftVerifyKey,
        email_template: type[EmailTemplate],
    ) -> None:
        linked_obj = LinkedObject.with_context(request, context=context)
        notification = CreateNotification(
            subject=message_subject,
            from_user_verify_key=context.credentials,
            to_user_verify_key=to_user_verify_key,
            linked_obj=linked_obj,
            notifier_types=[NOTIFIERS.EMAIL],
            email_template=email_template,
        )
        context.server.services.notification.send(
            context=context, notification=notification
        )

    @service_method(path="request.undo", name="undo", unwrap_on_success=False)
    def undo(self, context: AuthedServiceContext, uid: UID, reason: str) -> SyftSuccess:
        request: Request = self.stash.get_by_uid(
            credentials=context.credentials, uid=uid
        ).unwrap()

        context.extra_kwargs["reason"] = reason
        request.undo(context=context)

        self._send_email_notification(
            context=context,
            message_subject=f"Your request ({str(uid)[:4]}) has been denied. ",
            request=request,
            to_user_verify_key=request.requesting_user_verify_key,
            email_template=RequestUpdateEmailTemplate,
        )

        return SyftSuccess(message=f"Request {uid} successfully denied!")

    def save(self, context: AuthedServiceContext, request: Request) -> Request:
        return self.stash.update(context.credentials, request).unwrap()

    @service_method(
        path="request.delete_by_uid", name="delete_by_uid", unwrap_on_success=False
    )
    def delete_by_uid(self, context: AuthedServiceContext, uid: UID) -> SyftSuccess:
        """Delete the request with the given uid."""
        self.stash.delete_by_uid(context.credentials, uid).unwrap()
        return SyftSuccess(message=f"Request with id {uid} deleted.", value=uid)

    @service_method(
        path="request.set_tags",
        name="set_tags",
        roles=ADMIN_ROLE_LEVEL,
    )
    def set_tags(
        self,
        context: AuthedServiceContext,
        request: Request,
        tags: list[str],
    ) -> Request:
        request = self.stash.get_by_uid(context.credentials, request.id).unwrap()
        request.tags = tags
        return self.save(context, request)

    @service_method(path="request.get_by_usercode_id", name="get_by_usercode_id")
    def get_by_usercode_id(
        self, context: AuthedServiceContext, usercode_id: UID
    ) -> list[Request]:
        return self.stash.get_by_usercode_id(context.credentials, usercode_id).unwrap()


TYPE_TO_SERVICE[Request] = RequestService
SERVICE_TO_TYPES[RequestService].update({Request})



# stdlib
from typing import Any

# relative
from ...types.errors import SyftException
from ...types.result import as_result
from ...util.util import get_mb_serialized_size
from ..metadata.server_metadata import ServerMetadata
from ..metadata.server_metadata import ServerMetadataJSON


def min_size_for_blob_storage_upload(
    metadata: ServerMetadata | ServerMetadataJSON,
) -> int:
    return metadata.min_size_blob_storage_mb


@as_result(SyftException)
def can_upload_to_blob_storage(
    data: Any, metadata: ServerMetadata | ServerMetadataJSON
) -> bool:
    try:
        return get_mb_serialized_size(data) >= min_size_for_blob_storage_upload(
            metadata
        )
    except TypeError as exc:
        raise SyftException.from_exception(exc, public_message=str(exc))



# stdlib
import base64
from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
import json
import typing
from typing import Any
from typing import Generic
from typing import TypeVar
from typing import Union
from typing import get_args
from typing import get_origin

# third party
import pydantic

# syft absolute
import syft as sy

# relative
from ..server.credentials import SyftSigningKey
from ..server.credentials import SyftVerifyKey
from ..types.datetime import DateTime
from ..types.syft_object import BaseDateTime
from ..types.syft_object_registry import SyftObjectRegistry
from ..types.uid import LineageID
from ..types.uid import UID
from .recursive import DEFAULT_EXCLUDE_ATTRS

T = TypeVar("T")

JSON_CANONICAL_NAME_FIELD = "__canonical_name__"
JSON_VERSION_FIELD = "__version__"
JSON_DATA_FIELD = "data"

JsonPrimitive = str | int | float | bool | None
Json = JsonPrimitive | list["Json"] | dict[str, "Json"]


def _noop_fn(obj: Any) -> Any:
    return obj


@dataclass
class JSONSerde(Generic[T]):
    klass: type[T]
    serialize_fn: Callable[[T], Json]
    deserialize_fn: Callable[[Json], T]

    def serialize(self, obj: T) -> Json:
        return self.serialize_fn(obj)

    def deserialize(self, obj: Json) -> T:
        return self.deserialize_fn(obj)


JSON_SERDE_REGISTRY: dict[type[T], JSONSerde[T]] = {}


def register_json_serde(
    type_: type[T],
    serialize: Callable[[T], Json] | None = None,
    deserialize: Callable[[Json], T] | None = None,
) -> None:
    if type_ in JSON_SERDE_REGISTRY:
        raise ValueError(f"Type {type_} is already registered")

    if serialize is None:
        serialize = _noop_fn

    if deserialize is None:
        deserialize = _noop_fn

    JSON_SERDE_REGISTRY[type_] = JSONSerde(
        klass=type_,
        serialize_fn=serialize,
        deserialize_fn=deserialize,
    )


# Standard JSON primitives
register_json_serde(int)
register_json_serde(str)
register_json_serde(bool)
register_json_serde(float)
register_json_serde(type(None))
register_json_serde(pydantic.EmailStr)

# Syft primitives
register_json_serde(UID, lambda uid: uid.no_dash, lambda s: UID(s))
register_json_serde(LineageID, lambda uid: uid.no_dash, lambda s: LineageID(s))
register_json_serde(
    DateTime, lambda dt: dt.utc_timestamp, lambda f: DateTime(utc_timestamp=f)
)
register_json_serde(
    BaseDateTime, lambda dt: dt.utc_timestamp, lambda f: BaseDateTime(utc_timestamp=f)
)
register_json_serde(SyftVerifyKey, lambda key: str(key), SyftVerifyKey.from_string)
register_json_serde(SyftSigningKey, lambda key: str(key), SyftSigningKey.from_string)


def _validate_json(value: T) -> T:
    # Throws TypeError if value is not JSON-serializable
    json.dumps(value)
    return value


def _is_optional_annotation(annotation: Any) -> bool:
    try:
        return annotation | None == annotation
    except TypeError:
        return False


def _is_annotated_type(annotation: Any) -> bool:
    return get_origin(annotation) == typing.Annotated


def _unwrap_optional_annotation(annotation: Any) -> Any:
    """Return the type anntation with None type removed, if it is present.

    Args:
        annotation (Any): type annotation

    Returns:
        Any: type annotation without None type
    """
    if _is_optional_annotation(annotation):
        args = get_args(annotation)
        return Union[tuple(arg for arg in args if arg is not type(None))]  # noqa
    return annotation


def _unwrap_annotated(annotation: Any) -> Any:
    # Convert Annotated[T, ...] to T
    return get_args(annotation)[0]


def _unwrap_type_annotation(annotation: Any) -> Any:
    """
    recursively unwrap type annotations, removing Annotated and Optional types
    """
    if _is_annotated_type(annotation):
        res = _unwrap_annotated(annotation)
        return _unwrap_type_annotation(res)
    elif _is_optional_annotation(annotation):
        res = _unwrap_optional_annotation(annotation)
        return _unwrap_type_annotation(res)
    return annotation


def _annotation_issubclass(annotation: Any, cls: type) -> bool:
    # issubclass throws TypeError if annotation is not a valid type (eg Union)
    try:
        return issubclass(annotation, cls)
    except TypeError:
        return False


def _serialize_pydantic_to_json(obj: pydantic.BaseModel) -> dict[str, Json]:
    canonical_name, version = SyftObjectRegistry.get_canonical_name_version(obj)
    serde_attributes = SyftObjectRegistry.get_serde_properties(canonical_name, version)
    exclude_attrs = serde_attributes[4]

    result: dict[str, Json] = {
        JSON_CANONICAL_NAME_FIELD: canonical_name,
        JSON_VERSION_FIELD: version,
    }

    all_exclude_attrs = set(exclude_attrs) | DEFAULT_EXCLUDE_ATTRS

    for key, type_ in obj.model_fields.items():
        if key in all_exclude_attrs:
            continue
        result[key] = serialize_json(getattr(obj, key), type_.annotation)

    result = _add_searchable_and_unique_attrs(obj, result, raise_errors=False)

    return result


def get_property_return_type(obj: Any, attr_name: str) -> Any:
    """
    Get the return type annotation of a @property.
    """
    cls = type(obj)
    attr = getattr(cls, attr_name, None)

    if isinstance(attr, property):
        return attr.fget.__annotations__.get("return", None)

    return None


def _add_searchable_and_unique_attrs(
    obj: pydantic.BaseModel, obj_dict: dict[str, Json], raise_errors: bool = True
) -> dict[str, Json]:
    """
    Add searchable attrs and unique attrs to the serialized object dict, if they are not already present.
    Needed for adding non-field attributes (like @property)

    Args:
        obj (pydantic.BaseModel): Object to serialize.
        obj_dict (dict[str, Json]): Serialized object dict. Should contain the object's fields.
        raise_errors (bool, optional): Raise errors if an attribute cannot be accessed.
            If False, the attribute will be skipped. Defaults to True.

    Raises:
        Exception: Any exception raised when accessing an attribute.

    Returns:
        dict[str, Json]: Serialized object dict including searchable attributes.
    """
    searchable_attrs: list[str] = getattr(obj, "__attr_searchable__", [])
    unique_attrs: list[str] = getattr(obj, "__attr_unique__", [])

    attrs_to_add = set(searchable_attrs) | set(unique_attrs)
    for attr in attrs_to_add:
        if attr not in obj_dict:
            try:
                value = getattr(obj, attr)
            except Exception as e:
                if raise_errors:
                    raise e
                else:
                    continue
            property_annotation = get_property_return_type(obj, attr)
            obj_dict[attr] = serialize_json(
                value, validate=False, annotation=property_annotation
            )

    return obj_dict


def _deserialize_pydantic_from_json(
    obj_dict: dict[str, Json],
) -> pydantic.BaseModel:
    try:
        canonical_name = obj_dict[JSON_CANONICAL_NAME_FIELD]
        version = obj_dict[JSON_VERSION_FIELD]
        obj_type = SyftObjectRegistry.get_serde_class(canonical_name, version)

        result = {}
        for key, type_ in obj_type.model_fields.items():
            if key not in obj_dict:
                continue
            result[key] = deserialize_json(obj_dict[key], type_.annotation)

        return obj_type.model_validate(result)
    except Exception as e:
        print(f"Failed to deserialize Pydantic model: {e}")
        print(json.dumps(obj_dict, indent=2))
        raise ValueError(f"Failed to deserialize Pydantic model: {e}")


def _is_serializable_iterable(annotation: Any) -> bool:
    # we can only serialize typed iterables without Union/Any
    # NOTE optional is allowed

    # 1. check if it is an iterable
    if get_origin(annotation) not in {list, tuple, set, frozenset}:
        return False

    # 2. check if iterable annotation is serializable
    args = get_args(annotation)
    if len(args) != 1:
        return False

    inner_type = _unwrap_type_annotation(args[0])
    return inner_type in JSON_SERDE_REGISTRY or _annotation_issubclass(
        inner_type, pydantic.BaseModel
    )


def _serialize_iterable_to_json(value: Any, annotation: Any) -> Json:
    # No need to validate in recursive calls
    return [serialize_json(v, validate=False) for v in value]


def _deserialize_iterable_from_json(value: Json, annotation: Any) -> Any:
    if not isinstance(value, list):
        raise ValueError(f"Cannot deserialize {type(value)} to {annotation}")

    annotation = _unwrap_type_annotation(annotation)

    if not _is_serializable_iterable(annotation):
        raise ValueError(f"Cannot deserialize {annotation} from JSON")

    inner_type = _unwrap_type_annotation(get_args(annotation)[0])
    return [deserialize_json(v, inner_type) for v in value]


def _is_serializable_mapping(annotation: Any) -> bool:
    """
    Mapping is serializable if:
    - it is a dict
    - the key type is str
    - the value type is serializable and not a Union
    """
    if get_origin(annotation) != dict:
        return False

    args = get_args(annotation)
    if len(args) != 2:
        return False

    key_type, value_type = args
    # JSON only allows string keys
    if not isinstance(key_type, str):
        return False

    # check if value type is serializable
    value_type = _unwrap_type_annotation(value_type)
    return value_type in JSON_SERDE_REGISTRY or _annotation_issubclass(
        value_type, pydantic.BaseModel
    )


def _serialize_mapping_to_json(value: Any, annotation: Any) -> Json:
    _, value_type = get_args(annotation)
    # No need to validate in recursive calls
    return {k: serialize_json(v, value_type, validate=False) for k, v in value.items()}


def _deserialize_mapping_from_json(value: Json, annotation: Any) -> Any:
    if not isinstance(value, dict):
        raise ValueError(f"Cannot deserialize {type(value)} to {annotation}")

    annotation = _unwrap_type_annotation(annotation)

    if not _is_serializable_mapping(annotation):
        raise ValueError(f"Cannot deserialize {annotation} from JSON")

    _, value_type = get_args(annotation)
    return {k: deserialize_json(v, value_type) for k, v in value.items()}


def _serialize_to_json_bytes(obj: Any) -> str:
    obj_bytes = sy.serialize(obj, to_bytes=True)
    return base64.b64encode(obj_bytes).decode("utf-8")


def _deserialize_from_json_bytes(obj: str) -> Any:
    obj_bytes = base64.b64decode(obj)
    return sy.deserialize(obj_bytes, from_bytes=True)


def serialize_json(value: Any, annotation: Any = None, validate: bool = True) -> Json:
    """
    Serialize a value to a JSON-serializable object, using the schema defined by the
    provided annotation.

    Serialization is always done according to the annotation, as the same annotation
    is used for deserialization. If the annotation is not provided or is ambiguous,
    the JSON serialization will fall back to serializing bytes. Examples:
    - int, `list[int]` are strictly typed
    - `str | int`, `list`, `list[str | int]`, `list[Any]` are ambiguous and serialized to bytes
    - Optional types (like int | None) are serialized to the not-None type

    The function chooses the appropriate serialization method in the following order:
    1. Method registered in `JSON_SERDE_REGISTRY` for the annotation type.
    2. Pydantic model serialization, including all `SyftObjects`.
    3. Iterable serialization, if the annotation is a strict iterable (e.g., `list[int]`).
    4. Mapping serialization, if the annotation is a strictly typed mapping with string keys.
    5. Serialize the object to bytes and encode it as base64.

    Args:
        value (Any): Value to serialize.
        annotation (Any, optional): Type annotation for the value. Defaults to None.

    Returns:
        Json: JSON-serializable object.
    """
    if annotation is None:
        annotation = type(value)

    if value is None:
        return None

    # Remove None type from annotation if it is present.
    annotation = _unwrap_type_annotation(annotation)

    if annotation in JSON_SERDE_REGISTRY:
        result = JSON_SERDE_REGISTRY[annotation].serialize(value)
    elif _annotation_issubclass(annotation, pydantic.BaseModel):
        result = _serialize_pydantic_to_json(value)
    elif _annotation_issubclass(annotation, Enum):
        result = value.name

    # JSON recursive types
    # only strictly annotated iterables and mappings are supported
    # example: list[int] is supported, but not list[int | str]
    elif _is_serializable_iterable(annotation):
        result = _serialize_iterable_to_json(value, annotation)
    elif _is_serializable_mapping(annotation):
        result = _serialize_mapping_to_json(value, annotation)
    else:
        result = _serialize_to_json_bytes(value)

    if validate:
        _validate_json(result)

    return result


def deserialize_json(value: Json, annotation: Any = None) -> Any:
    """Deserialize a JSON-serializable object to a value, using the schema defined by the
    provided annotation. Inverse of `serialize_json`.

    Args:
        value (Json): JSON-serializable object.
        annotation (Any): Type annotation for the value.

    Returns:
        Any: Deserialized value.
    """
    if (
        isinstance(value, dict)
        and JSON_CANONICAL_NAME_FIELD in value
        and JSON_VERSION_FIELD in value
    ):
        return _deserialize_pydantic_from_json(value)

    if value is None:
        return None

    # Remove None type from annotation if it is present.
    if annotation is None:
        raise ValueError("Annotation is required for deserialization")

    annotation = _unwrap_type_annotation(annotation)

    if annotation in JSON_SERDE_REGISTRY:
        return JSON_SERDE_REGISTRY[annotation].deserialize(value)
    elif _annotation_issubclass(annotation, pydantic.BaseModel):
        return _deserialize_pydantic_from_json(value)
    elif _annotation_issubclass(annotation, Enum):
        return annotation[value]
    elif isinstance(value, list):
        return _deserialize_iterable_from_json(value, annotation)
    elif isinstance(value, dict):
        return _deserialize_mapping_from_json(value, annotation)
    elif isinstance(value, str):
        return _deserialize_from_json_bytes(value)
    else:
        raise ValueError(f"Cannot deserialize {value} to {annotation}")


def is_json_primitive(value: Any) -> bool:
    serialized = serialize_json(value, validate=False)
    return isinstance(serialized, JsonPrimitive)  # type: ignore



# stdlib
import contextlib
from typing import Any
from typing import cast

# third party
import docker
from docker.models.containers import Container

# relative
from ...custom_worker.k8s import IN_KUBERNETES
from ...custom_worker.k8s import PodStatus
from ...custom_worker.runner_k8s import KubernetesRunner
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.db import DBManager
from ...store.document_store_errors import StashException
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.uid import UID
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import AuthedServiceContext
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import DATA_OWNER_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from .utils import DEFAULT_WORKER_POOL_NAME
from .utils import _get_healthcheck_based_on_status
from .utils import map_pod_to_worker_status
from .worker_pool import ContainerSpawnStatus
from .worker_pool import SyftWorker
from .worker_pool import WorkerHealth
from .worker_pool import WorkerStatus
from .worker_pool import _get_worker_container
from .worker_pool import _get_worker_container_status
from .worker_stash import WorkerStash


@serializable(canonical_name="WorkerService", version=1)
class WorkerService(AbstractService):
    stash: WorkerStash

    def __init__(self, store: DBManager) -> None:
        self.stash = WorkerStash(store=store)

    @service_method(
        path="worker.start_workers",
        name="start_workers",
        roles=ADMIN_ROLE_LEVEL,
    )
    def start_workers(
        self, context: AuthedServiceContext, n: int = 1
    ) -> list[ContainerSpawnStatus]:
        """Add a Container Image."""

        return context.server.services.syft_worker_pool.add_workers(
            context, number=n, pool_name=DEFAULT_WORKER_POOL_NAME
        )

    @service_method(
        path="worker.get_all", name="get_all", roles=DATA_SCIENTIST_ROLE_LEVEL
    )
    def list(self, context: AuthedServiceContext) -> list[SyftWorker]:
        """List all the workers."""
        workers = self.stash.get_all(context.credentials).unwrap()

        if context.server is not None and context.server.in_memory_workers:
            return workers
        else:
            # If container workers, check their statuses
            workers = refresh_worker_status(
                workers, self.stash, context.as_root_context().credentials
            ).unwrap()
        return workers

    @service_method(
        path="worker.status", name="status", roles=DATA_SCIENTIST_ROLE_LEVEL
    )
    def status(
        self,
        context: AuthedServiceContext,
        uid: UID,
    ) -> tuple[WorkerStatus, WorkerHealth | None]:
        result = self.get(context=context, uid=uid)
        return result.status, result.healthcheck

    @service_method(
        path="worker.get",
        name="get",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get(self, context: AuthedServiceContext, uid: UID) -> SyftWorker:
        worker = self._get_worker(context=context, uid=uid).unwrap()

        if context.server is not None and context.server.in_memory_workers:
            return worker
        else:
            workers = refresh_worker_status(
                [worker], self.stash, context.as_root_context().credentials
            ).unwrap()
            return workers[0]

    @service_method(
        path="worker.logs",
        name="logs",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def logs(
        self,
        context: AuthedServiceContext,
        uid: UID,
        raw: bool = False,
    ) -> bytes | str:
        worker = self._get_worker(context=context, uid=uid).unwrap()

        if context.server is not None and context.server.in_memory_workers:
            logs = b"Logs not implemented for In Memory Workers"
        elif IN_KUBERNETES:
            runner = KubernetesRunner()
            return runner.get_pod_logs(pod_name=worker.name)
        else:
            with contextlib.closing(docker.from_env()) as client:
                docker_container = _get_worker_container(client, worker).unwrap()
                try:
                    logs = cast(bytes, docker_container.logs())
                except docker.errors.APIError as e:
                    raise SyftException(
                        public_message=f"Failed to get worker {worker.id} container logs. Error {e}"
                    )

        return logs if raw else logs.decode(errors="ignore")

    def _delete(
        self, context: AuthedServiceContext, worker: SyftWorker, force: bool = False
    ) -> SyftSuccess:
        uid = worker.id
        if force and worker.job_id is not None:
            context.server.services.job.kill(context=context, id=worker.job_id)

        worker_pool_stash = context.server.services.syft_worker_pool.stash
        worker_pool = worker_pool_stash.get_by_name(
            credentials=context.credentials, pool_name=worker.worker_pool_name
        ).unwrap()

        if IN_KUBERNETES:
            # Kubernetes will only restart the worker NOT REMOVE IT
            runner = KubernetesRunner()
            runner.delete_pod(pod_name=worker.name)
            return SyftSuccess(
                # pod deletion is not supported in Kubernetes, removing and recreating the pod.
                message=(
                    "Worker deletion is not supported in Kubernetes. "
                    f"Removing and re-creating worker id={worker.id}"
                )
            )
        elif not context.server.in_memory_workers:
            # delete the worker using docker client sdk
            with contextlib.closing(docker.from_env()) as client:
                docker_container = _get_worker_container(client, worker).unwrap()
                _stop_worker_container(worker, docker_container, force=force).unwrap()
        else:
            # kill the in memory worker thread
            context.server.remove_consumer_with_id(syft_worker_id=worker.id)

        # remove the worker from the pool
        try:
            worker_linked_object = next(
                obj for obj in worker_pool.worker_list if obj.object_uid == uid
            )
            worker_pool.worker_list.remove(worker_linked_object)
        except StopIteration:
            pass

        # Delete worker from worker stash
        self.stash.delete_by_uid(credentials=context.credentials, uid=uid).unwrap()

        # Update worker pool
        worker_pool_stash.update(context.credentials, obj=worker_pool).unwrap()

        return SyftSuccess(
            message=f"Worker with id: {uid} deleted successfully from pool: {worker_pool.name}"
        )

    @service_method(
        path="worker.delete",
        name="delete",
        roles=DATA_OWNER_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def delete(
        self,
        context: AuthedServiceContext,
        uid: UID,
        force: bool = False,
    ) -> SyftSuccess:
        worker = self._get_worker(context=context, uid=uid).unwrap()
        worker.to_be_deleted = True

        self.stash.update(context.credentials, worker).unwrap()
        if not force:
            # relative
            return SyftSuccess(message=f"Worker {uid} has been marked for deletion.")

        return self._delete(context, worker, force=True)

    @as_result(SyftException, StashException)
    def _get_worker(self, context: AuthedServiceContext, uid: UID) -> SyftWorker:
        return self.stash.get_by_uid(credentials=context.credentials, uid=uid).unwrap()


@as_result(SyftException)
def refresh_worker_status(
    workers: list[SyftWorker],
    worker_stash: WorkerStash,
    credentials: SyftVerifyKey,
) -> list[SyftWorker]:
    if IN_KUBERNETES:
        workers = refresh_status_kubernetes(workers).unwrap()
    else:
        workers = refresh_status_docker(workers).unwrap()

    for worker in workers:
        worker_stash.update(
            credentials=credentials,
            obj=worker,
        ).unwrap()

    return workers


@as_result(SyftException)
def refresh_status_kubernetes(workers: list[SyftWorker]) -> list[SyftWorker]:
    updated_workers = []
    runner = KubernetesRunner()
    for worker in workers:
        status: PodStatus | WorkerStatus | None = runner.get_pod_status(pod=worker.name)
        if not status:
            worker.status = WorkerStatus.STOPPED
            worker.healthcheck = WorkerHealth.UNHEALTHY
        else:
            status, health, _ = map_pod_to_worker_status(status)
            worker.status = status
            worker.healthcheck = health
            updated_workers.append(worker)

    return updated_workers


@as_result(SyftException)
def refresh_status_docker(workers: list[SyftWorker]) -> list[SyftWorker]:
    updated_workers = []
    with contextlib.closing(docker.from_env()) as client:
        for worker in workers:
            status = _get_worker_container_status(client, worker).unwrap()
            worker.status = status
            worker.healthcheck = _get_healthcheck_based_on_status(status=status)
            updated_workers.append(worker)
    return updated_workers


@as_result(SyftException)
def _stop_worker_container(
    worker: SyftWorker,
    container: Container,
    force: bool,
) -> None:
    try:
        # stop the container
        container.stop()
        # Remove the container and its volumes
        _remove_worker_container(container, force=force, v=True)
        return None
    except Exception as e:
        raise SyftException(
            public_message=f"Failed to delete worker with id: {worker.id}. Error: {e}"
        )


def _remove_worker_container(container: Container, **kwargs: Any) -> None:
    try:
        container.remove(**kwargs)
    except docker.errors.NotFound:
        return
    except docker.errors.APIError as e:
        if "removal of container" in str(e) and "is already in progress" in str(e):
            # If the container is already being removed, ignore the error
            return



# stdlib
from typing import Any

# third party
from typing_extensions import Self

# relative
from ..abstract_server import AbstractServer
from ..abstract_server import ServerSideType
from ..server.credentials import SyftVerifyKey
from ..server.credentials import UserLoginCredentials
from ..types.syft_object import Context
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import SyftBaseObject
from ..types.syft_object import SyftObject
from ..types.uid import UID
from .user.user_roles import ROLE_TO_CAPABILITIES
from .user.user_roles import ServiceRole
from .user.user_roles import ServiceRoleCapability


class ServerServiceContext(Context, SyftObject):
    __canonical_name__ = "ServerServiceContext"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID | None = None  # type: ignore[assignment]
    server: AbstractServer


class AuthedServiceContext(ServerServiceContext):
    __canonical_name__ = "AuthedServiceContext"
    __version__ = SYFT_OBJECT_VERSION_1

    credentials: SyftVerifyKey
    role: ServiceRole = ServiceRole.NONE
    job_id: UID | None = None
    extra_kwargs: dict = {}
    has_execute_permissions: bool = False
    is_blocking_api_call: bool = False
    client_warnings: list[str] = []

    @property
    def dev_mode(self) -> Any:
        return self.server.dev_mode  # type: ignore

    def add_warning(self, message: str) -> None:
        self.client_warnings.append(message)

    def capabilities(self) -> list[ServiceRoleCapability]:
        return ROLE_TO_CAPABILITIES.get(self.role, [])

    def with_credentials(self, credentials: SyftVerifyKey, role: ServiceRole) -> Self:
        return AuthedServiceContext(
            credentials=credentials, role=role, server=self.server
        )

    @property
    def is_l0_lowside(self) -> bool:
        """Returns True if this is a low side of a Level 0 deployment"""
        return self.server.server_side_type == ServerSideType.LOW_SIDE

    @property
    def server_allows_execution_for_ds(self) -> bool:
        """Returns True if this is a low side of a Level 0 deployment"""
        return not self.is_l0_lowside

    def as_root_context(self) -> Self:
        return AuthedServiceContext(
            credentials=self.server.verify_key,
            role=ServiceRole.ADMIN,
            server=self.server,
        )

    @property
    def job(self):  # type: ignore
        # TODO: fix the mypy issue. The return type is Optional[Job].
        # but we can't import Job since it's a circular import
        if self.job_id is None:
            return None
        return self.server.job_stash.get_by_uid(
            self.credentials, self.job_id
        ).ok()  # if this fails, it will return None


class UnauthedServiceContext(ServerServiceContext):
    __canonical_name__ = "UnauthedServiceContext"
    __version__ = SYFT_OBJECT_VERSION_1

    server: AbstractServer
    login_credentials: UserLoginCredentials | None = None
    role: ServiceRole = ServiceRole.NONE


class ChangeContext(SyftBaseObject):
    __canonical_name__ = "ChangeContext"
    __version__ = SYFT_OBJECT_VERSION_1

    server: AbstractServer
    approving_user_credentials: SyftVerifyKey | None = None
    requesting_user_credentials: SyftVerifyKey | None = None
    extra_kwargs: dict = {}

    @classmethod
    def from_service(cls, context: AuthedServiceContext) -> Self:
        return cls(
            server=context.server,
            approving_user_credentials=context.credentials,
            extra_kwargs=context.extra_kwargs,
        )

    def to_service_ctx(self) -> AuthedServiceContext:
        return AuthedServiceContext(
            server=self.server,
            credentials=self.approving_user_credentials,
            extra_kwargs=self.extra_kwargs,
        )



# stdlib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import logging
import smtplib

# third party
from pydantic import BaseModel

# relative
from ...types.errors import SyftException
from ...types.server_url import ServerURL

SOCKET_TIMEOUT = 5  # seconds

logger = logging.getLogger(__name__)


class SMTPClient(BaseModel):
    server: str
    port: int
    password: str | None = None
    username: str | None = None

    def send(self, sender: str, receiver: list[str], subject: str, body: str) -> None:
        if not (subject and body and receiver):
            raise ValueError("Subject, body, and recipient email(s) are required")

        msg = MIMEMultipart("alternative")
        msg["From"] = sender
        msg["To"] = ", ".join(receiver)
        msg["Subject"] = subject
        msg.attach(MIMEText(body, "html"))

        mail_url = ServerURL.from_url(f"smtp://{self.server}:{self.port}")
        mail_url = mail_url.as_container_host()
        try:
            with smtplib.SMTP(
                mail_url.host_or_ip, mail_url.port, timeout=SOCKET_TIMEOUT
            ) as server:
                server.ehlo()
                if server.has_extn("STARTTLS"):
                    server.starttls()
                    server.ehlo()
                if self.username and self.password:
                    server.login(self.username, self.password)
                text = msg.as_string()
                server.sendmail(sender, ", ".join(receiver), text)
                return None
        except Exception as e:
            logger.error(f"Unable to send email. {e}")
            raise SyftException(
                public_message="Oops! Something went wrong while trying to send an email."
            )

    @classmethod
    def check_credentials(
        cls, server: str, port: int, username: str, password: str
    ) -> bool:
        """Check if the credentials are valid.

        Returns:
            bool: True if the credentials are valid, False otherwise.
        """
        try:
            mail_url = ServerURL.from_url(f"smtp://{server}:{port}")
            mail_url = mail_url.as_container_host()

            print(f"> Validating SMTP settings: {mail_url}")
            with smtplib.SMTP(
                mail_url.host_or_ip, mail_url.port, timeout=SOCKET_TIMEOUT
            ) as smtp_server:
                smtp_server.ehlo()
                if smtp_server.has_extn("STARTTLS"):
                    smtp_server.starttls()
                    smtp_server.ehlo()
                smtp_server.login(username, password)
                return True
        except Exception as e:
            message = f"SMTP check_credentials failed. {e}"
            print(message)
            logger.error(message)
            raise SyftException(public_message=str(e))



# stdlib
import logging

# relative
from ...abstract_server import AbstractServer
from .user import User
from .user import UserCreate
from .user_roles import ServiceRole

logger = logging.getLogger(__name__)


def create_root_admin_if_not_exists(
    name: str,
    email: str,
    password: str,
    server: AbstractServer,
) -> User | None:
    """
    If no root admin exists:
    - all exists checks on the user stash will fail, as we cannot get the role for the admin to check if it exists
    - result: a new admin is always created

    If a root admin exists with a different email:
    - cause: DEFAULT_USER_EMAIL env variable is set to a different email than the root admin in the db
    - verify_key_exists will return True
    - result: no new admin is created, as the server already has a root admin
    """
    user_stash = server.services.user.stash

    email_exists = user_stash.email_exists(email=email).unwrap()
    if email_exists:
        logger.debug("Admin not created, a user with this email already exists")
        return None

    verify_key_exists = user_stash.verify_key_exists(server.verify_key).unwrap()
    if verify_key_exists:
        logger.debug("Admin not created, this server already has a root admin")
        return None

    create_user = UserCreate(
        name=name,
        email=email,
        password=password,
        password_verify=password,
        role=ServiceRole.ADMIN,
    )

    # New User Initialization
    # 🟡 TODO: change later but for now this gives the main user super user automatically
    user = create_user.to(User)
    user.signing_key = server.signing_key
    user.verify_key = server.verify_key

    new_user = user_stash.set(
        credentials=server.verify_key,
        obj=user,
        ignore_duplicates=False,
    ).unwrap()

    logger.debug(f"Created admin {new_user.email}")

    return new_user



# relative
from ...types.errors import SyftException


class UserError(SyftException):
    public_message = "UserError. Please contact the admin."


class UserCreateError(UserError):
    public_message = "Failed to create user."


class UserDeleteError(UserError):
    public_message = "Failed to delete user."


class UserUpdateError(UserError):
    public_message = "Failed to update user."


class UserPasswordMismatchError(UserError):
    public_message = "Passwords do not match!"


class UserInvalidEmailError(UserError):
    public_message = "Invalid email address."


class UserSearchBadParamsError(UserError): ...


# public_message = (
#     f"Invalid Search parameters. Allowed params: "
#     f"{list(UserSearch.model_fields.keys())}"
# )


class UserPermissionError(UserError):
    public_message = "You are not permitted to perform this action."


class UserExchangeCredentials(UserError):
    public_message = "Invalid credential exchange. Please contact the admin."


class UserEnclaveAdminLoginError(UserError):
    public_message = (
        "Admins are not allowed to login to Enclaves.\n"
        "Kindly register a new data scientist account via `client.register`."
    )



# stdlib
from collections import OrderedDict
import hashlib
from inspect import Signature
import json
import os
from pathlib import Path

# relative
from ...service.service import ServiceConfigRegistry
from ...service.warnings import APIEndpointWarning
from ..util import get_root_data_path
from ..util import str_to_bool
from .json_diff import json_diff

API_SPEC_JSON_FILENAME = "syft_api_spec.json"
API_DIFF_JSON_FILENAME = "syft_api_diff.json"


def api_snapshot_dir() -> Path:
    """A helper function to get the path of the API snapshot directory."""
    return Path(os.path.abspath(str(Path(__file__).parent)))


class SyftAPISnapshot:
    def __init__(
        self,
        filename: str = API_SPEC_JSON_FILENAME,
        stable_release: bool = False,
    ) -> None:
        """
        Initialize the SyftAPISnapshot object.

        Args:
            filename (str): The name of the JSON file to load the API snapshot from.
                            Defaults to API_SPEC_JSON_FILENAME.
        """

        filename = self.get_filename(filename, stable_release)
        self.file_path = api_snapshot_dir() / filename
        self.history = self.load_map()
        self.state = self.build_map()

    def get_filename(self, filename: str, stable_release: bool) -> str:
        """
        Get the modified filename based on the stable_release flag.

        Args:
            filename (str): The original filename.
            stable_release (bool): Flag indicating if it's a stable release.

        Returns:
            str: The modified filename.
        """
        if stable_release:
            return f"{filename.split('.')[0]}_stable.json"
        else:
            return f"{filename.split('.')[0]}_beta.json"

    @staticmethod
    def extract_service_name(path: str) -> str:
        """
        Extract the service name from the given path.

        Args:
            path (str): The path of the service.

        Returns:
            str: The extracted service name.
        """
        return path.split(".")[0].capitalize()

    @staticmethod
    def extract_arguments(signature: Signature) -> dict:
        """
        Extract the arguments from the given signature.

        Args:
            signature (Signature): The signature object.

        Returns:
            dict: The extracted arguments as a dictionary.
        """
        signature_kwargs = {
            f"{v.name}": f"{v.annotation}" for k, v in signature.parameters.items()
        }
        return OrderedDict(sorted(signature_kwargs.items()))

    @staticmethod
    def get_role_level(roles: list) -> str:
        """
        Get the role level from the given list of roles.

        Args:
            roles (list): The list of roles.

        Returns:
            str: The role level.
        """
        return sorted(roles)[0].name + "_ROLE_LEVEL"

    @staticmethod
    def extract_warning_info(warning: APIEndpointWarning) -> dict | str:
        """
        Extract the warning information from the given APIEndpointWarning object.

        Args:
            warning (APIEndpointWarning): The APIEndpointWarning object.

        Returns:
            dict: The extracted warning information.
        """
        if not warning:
            return ""

        return {
            "name": f"{warning.__class__.__name__}",
            "confirmation": warning.confirmation,
            "enabled": warning.enabled,
        }

    @staticmethod
    def generate_hash(api_map: OrderedDict) -> str:
        """
        Generate a hash for the given API map.

        Args:
            api_map (OrderedDict): The API map.

        Returns:
            str: The generated hash.
        """
        return hashlib.sha256(json.dumps(api_map).encode()).hexdigest()

    def load_map(self) -> OrderedDict:
        """
        Load the API map from the JSON file.

        Returns:
            OrderedDict: The loaded API map.
        """

        if not self.file_path.exists():
            return OrderedDict()

        return OrderedDict(json.loads(self.file_path.read_text()))

    def build_map(self) -> OrderedDict:
        """
        Build the API map.

        Returns:
            OrderedDict: The built API map.
        """
        api_details = {}
        for (
            _,
            service_config,
        ) in ServiceConfigRegistry.__service_config_registry__.items():
            service_name = self.extract_service_name(service_config.private_path)
            warning = service_config.warning
            signature = service_config.signature
            role_level = self.get_role_level(service_config.roles)
            api_detail = {
                "public_path": service_config.public_path,
                "RBAC_permission": f"{role_level}",
                "signature": self.extract_arguments(service_config.signature),
                "return_type": f"{signature.return_annotation}" if signature else "",
                "warning": self.extract_warning_info(warning),
                # "unwrap_on_success": getattrservice_config.unwrap_on_success,
            }
            api_detail["hash"] = self.generate_hash(api_detail)
            api_details[f"{service_name}.{service_config.public_path}"] = OrderedDict(
                api_detail
            )

        api_details_ordered = OrderedDict(sorted(api_details.items()))
        return api_details_ordered

    def save_as_json(self) -> None:
        """
        Save the API map as a JSON file.
        """
        self.file_path.write_text(json.dumps(self.state, indent=2))

    def calc_diff(self, save: bool = False) -> dict:
        """
        Calculate the difference between the current API snapshot and the previous one.

        Args:
            save (bool): Whether to save the difference as a JSON file. Defaults to False.

        Returns:
            dict: The difference between the API snapshots.
        """
        diff = json_diff(self.history, self.state)
        if save:
            diff_file_path = get_root_data_path() / API_DIFF_JSON_FILENAME
            diff_file_path.write_text(json.dumps(diff, indent=2))

        return diff


def get_api_snapshot(stable_release: bool = False) -> SyftAPISnapshot:
    """
    Retrieves the API snapshot.
    """
    snapshot = SyftAPISnapshot(
        filename=API_SPEC_JSON_FILENAME,
        stable_release=stable_release,
    )
    return snapshot


def take_api_snapshot() -> SyftAPISnapshot:
    """
    Takes a stable release snapshot of the API and saves it as a JSON file.
    """

    # Get the stable_release flag from the environment variable
    stable_release = str_to_bool(os.environ.get("STABLE_RELEASE", "False"))

    snapshot = get_api_snapshot(stable_release=stable_release)
    snapshot.save_as_json()
    print("API snapshot saved at: ", snapshot.file_path)
    return snapshot


def show_api_diff() -> None:
    """
    Calculates the difference between the current API snapshot and the previous one,
    saves it as a JSON file, and returns the difference.
    """

    # Get the stable_release flag from the environment variable
    stable_release = str_to_bool(os.environ.get("STABLE_RELEASE", "False"))

    snapshot = get_api_snapshot(stable_release=stable_release)

    # Calculate the difference between the current API snapshot and the previous one
    diff = snapshot.calc_diff(save=True)
    print(json.dumps(diff, indent=2))
    print("Generated API diff file at: ", get_root_data_path() / API_DIFF_JSON_FILENAME)



# relative
class KeyValueBackingStore:
    pass


class KeyValueStorePartition:
    """Key-Value StorePartition
    Parameters:
        `settings`: PartitionSettings
            PySyft specific settings
        `store_config`: StoreConfig
            Backend specific configuration
    """

    pass



# relative
from ..types.errors import SyftException


class NotFoundException(SyftException):
    public_message = "Item not found."


class TooManyItemsFoundException(SyftException):
    public_message = "Too many items found."


class StashException(SyftException):
    public_message = "There was an error retrieving data. Contact your admin."


class UniqueConstraintException(StashException):
    public_message = "Another item with the same unique constraint already exists."


class ObjectCRUDPermissionException(SyftException):
    public_message = "You do not have permission to perform this action."


class ObjectExecutionPermissionException(SyftException):
    public_message = "You do not have permission to execute this action."



# syft absolute
import syft as sy


def make_submit_query(settings, worker_pool_name):
    updated_settings = {"user_code_worker": worker_pool_name} | settings

    @sy.api_endpoint(
        path="bigquery.submit_query",
        description="API endpoint that allows you to submit SQL queries to run on the private data.",
        worker_pool_name=worker_pool_name,
        settings=updated_settings,
    )
    def submit_query(
        context,
        func_name: str,
        query: str,
    ) -> str:
        # syft absolute
        import syft as sy

        @sy.syft_function(
            name=func_name,
            input_policy=sy.MixedInputPolicy(
                endpoint=sy.Constant(
                    val=context.user_client.api.services.bigquery.test_query
                ),
                query=sy.Constant(val=query),
                client=context.user_client,
            ),
            worker_pool_name=context.settings["user_code_worker"],
        )
        def execute_query(query: str, endpoint):
            res = endpoint(sql_query=query)
            return res

        request = context.user_client.code.request_code_execution(execute_query)
        context.admin_client.requests.set_tags(request, ["autosync"])

        return f"Query submitted {request}. Use `client.code.{func_name}()` to run your query"

    return submit_query



# stdlib
import threading
import time
from typing import Any

# third party
from pydantic import field_validator

# relative
from ...server.credentials import SyftVerifyKey
from ...types.base import SyftBaseModel
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.uid import UID
from ..worker.worker_pool import SyftWorker
from ..worker.worker_stash import WorkerStash

# Producer/Consumer heartbeat interval (in seconds)
HEARTBEAT_INTERVAL_SEC = 2

# Thread join timeout (in seconds)
THREAD_TIMEOUT_SEC = 5

# Max duration (in ms) to wait for ZMQ poller to return
ZMQ_POLLER_TIMEOUT_MSEC = 1000

# Duration (in seconds) after which a worker without a heartbeat will be marked as expired
WORKER_TIMEOUT_SEC = 60

# Duration (in seconds) after which producer without a heartbeat will be marked as expired
PRODUCER_TIMEOUT_SEC = 60

# Lock for working on ZMQ socket
ZMQ_SOCKET_LOCK = threading.Lock()

MAX_RECURSION_NESTED_ACTIONOBJECTS = 5


class ZMQHeader:
    """Enum for ZMQ headers"""

    W_WORKER = b"MDPW01"


class ZMQCommand:
    """Enum for ZMQ commands"""

    W_READY = b"0x01"
    W_REQUEST = b"0x02"
    W_REPLY = b"0x03"
    W_HEARTBEAT = b"0x04"
    W_DISCONNECT = b"0x05"


class Timeout:
    def __init__(self, offset_sec: float):
        self.__offset = float(offset_sec)
        self.__next_ts: float = 0.0

        self.reset()

    @property
    def next_ts(self) -> float:
        return self.__next_ts

    def reset(self) -> None:
        self.__next_ts = self.now() + self.__offset

    def has_expired(self) -> bool:
        return self.now() >= self.__next_ts

    @staticmethod
    def now() -> float:
        return time.time()


class Service:
    def __init__(self, name: str) -> None:
        self.name = name
        self.requests: list[bytes] = []
        self.waiting: list[Worker] = []  # List of waiting workers


class Worker(SyftBaseModel):
    address: bytes
    identity: bytes
    service: Service | None = None
    syft_worker_id: UID | None = None
    expiry_t: Timeout = Timeout(WORKER_TIMEOUT_SEC)

    @field_validator("syft_worker_id", mode="before")
    @classmethod
    def set_syft_worker_id(cls, v: Any) -> Any:
        if isinstance(v, str):
            return UID(v)
        return v

    def has_expired(self) -> bool:
        return self.expiry_t.has_expired()

    def get_expiry(self) -> float:
        return self.expiry_t.next_ts

    def reset_expiry(self) -> None:
        self.expiry_t.reset()

    @as_result(SyftException)
    def _syft_worker(
        self, stash: WorkerStash, credentials: SyftVerifyKey
    ) -> SyftWorker | None:
        return stash.get_by_uid(
            credentials=credentials, uid=self.syft_worker_id
        ).unwrap()

    def __str__(self) -> str:
        svc = self.service.name if self.service else None
        return (
            f"Worker(addr={self.address!r}, id={self.identity!r}, service={svc}, "
            f"syft_worker_id={self.syft_worker_id!r})"
        )



# relative
from ...serde.serializable import serializable
from ...store.db.stash import ObjectStash
from .log import SyftLog


@serializable(canonical_name="LogStash", version=1)
class LogStash(ObjectStash[SyftLog]):
    pass



# stdlib
import logging
from pathlib import Path
from typing import Generic
from typing import TypeVar
from urllib.parse import urlparse

# third party
from pydantic import BaseModel
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...types.uid import UID
from ...util.telemetry import instrument_sqlalchemny
from .schema import PostgresBase
from .schema import SQLiteBase

logger = logging.getLogger(__name__)
instrument_sqlalchemny()


@serializable(canonical_name="DBConfig", version=1)
class DBConfig(BaseModel):
    @property
    def connection_string(self) -> str:
        raise NotImplementedError("Subclasses must implement this method.")

    @classmethod
    def from_connection_string(cls, conn_str: str) -> "DBConfig":
        # relative
        from .postgres import PostgresDBConfig
        from .sqlite import SQLiteDBConfig

        parsed = urlparse(conn_str)
        if parsed.scheme == "postgresql":
            return PostgresDBConfig(
                host=parsed.hostname,
                port=parsed.port,
                user=parsed.username,
                password=parsed.password,
                database=parsed.path.lstrip("/"),
            )
        elif parsed.scheme == "sqlite":
            path = Path(parsed.path)
            return SQLiteDBConfig(path=path.parent, filename=path.name)
        else:
            raise ValueError(f"Unsupported database scheme: {parsed.scheme}")


ConfigT = TypeVar("ConfigT", bound=DBConfig)


class DBManager(Generic[ConfigT]):
    def __init__(
        self,
        config: ConfigT,
        server_uid: UID,
        root_verify_key: SyftVerifyKey,
    ) -> None:
        self.config = config
        self.root_verify_key = root_verify_key
        self.server_uid = server_uid
        self.engine = create_engine(
            config.connection_string,
            # json_serializer=dumps,
            # json_deserializer=loads,
        )
        logger.info(f"Connecting to {config.connection_string}")
        self.sessionmaker = sessionmaker(bind=self.engine)
        self.update_settings()
        logger.info(f"Successfully connected to {config.connection_string}")

    def update_settings(self) -> None:
        pass

    def init_tables(self, reset: bool = False) -> None:
        Base = SQLiteBase if self.engine.dialect.name == "sqlite" else PostgresBase

        with self.sessionmaker().begin() as _:
            if reset:
                Base.metadata.drop_all(bind=self.engine)
            Base.metadata.create_all(self.engine)



# stdlib

# third party

# relative
from ...client.api import ServerIdentity
from ...serde.serializable import serializable
from ...store.db.db import DBManager
from ...store.db.stash import ObjectStash
from ...types.syft_object import PartialSyftObject
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.uid import UID
from ..context import AuthedServiceContext
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from .user_code import ApprovalDecision
from .user_code import UserCodeStatusCollection


@serializable(canonical_name="StatusSQLStash", version=1)
class StatusStash(ObjectStash[UserCodeStatusCollection]):
    pass


class CodeStatusUpdate(PartialSyftObject):
    __canonical_name__ = "CodeStatusUpdate"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    decision: ApprovalDecision


@serializable(canonical_name="UserCodeStatusService", version=1)
class UserCodeStatusService(AbstractService):
    stash: StatusStash

    def __init__(self, store: DBManager):
        self.stash = StatusStash(store=store)

    @service_method(path="code_status.create", name="create", roles=ADMIN_ROLE_LEVEL)
    def create(
        self,
        context: AuthedServiceContext,
        status: UserCodeStatusCollection,
    ) -> UserCodeStatusCollection:
        res = self.stash.set(
            credentials=context.credentials,
            obj=status,
        ).unwrap()
        return res

    @service_method(
        path="code_status.update",
        name="update",
        roles=ADMIN_ROLE_LEVEL,
        autosplat=["code_update"],
        unwrap_on_success=False,
    )
    def update(
        self, context: AuthedServiceContext, code_update: CodeStatusUpdate
    ) -> SyftSuccess:
        existing_status = self.stash.get_by_uid(
            context.credentials, uid=code_update.id
        ).unwrap()
        server_identity = ServerIdentity.from_server(context.server)
        existing_status.status_dict[server_identity] = code_update.decision

        res = self.stash.update(context.credentials, existing_status).unwrap()
        return SyftSuccess(message="UserCode updated successfully", value=res)

    @service_method(
        path="code_status.get_by_uid", name="get_by_uid", roles=GUEST_ROLE_LEVEL
    )
    def get_status(
        self, context: AuthedServiceContext, uid: UID
    ) -> UserCodeStatusCollection:
        """Get the status of a user code item"""
        return self.stash.get_by_uid(context.credentials, uid=uid).unwrap()

    @service_method(path="code_status.get_all", name="get_all", roles=ADMIN_ROLE_LEVEL)
    def get_all(self, context: AuthedServiceContext) -> list[UserCodeStatusCollection]:
        """Get all user code item statuses"""
        return self.stash.get_all(context.credentials).unwrap()

    @service_method(
        path="code_status.remove",
        name="remove",
        roles=ADMIN_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def remove(self, context: AuthedServiceContext, uid: UID) -> SyftSuccess:
        """Remove a user code item status"""
        self.stash.delete_by_uid(context.credentials, uid=uid).unwrap()
        return SyftSuccess(message=f"{uid} successfully deleted", value=uid)


TYPE_TO_SERVICE[UserCodeStatusCollection] = UserCodeStatusService



# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftSigningKey
from ...server.credentials import SyftVerifyKey
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...types.result import as_result
from .user import User
from .user_roles import ServiceRole


@serializable(canonical_name="UserStashSQL", version=1)
class UserStash(ObjectStash[User]):
    @as_result(StashException, NotFoundException)
    def admin_user(self) -> User:
        # TODO: This returns only one user, the first user with the role ADMIN
        admin_credentials = self.root_verify_key
        return self.get_by_role(
            credentials=admin_credentials, role=ServiceRole.ADMIN
        ).unwrap()

    @as_result(StashException, NotFoundException)
    def get_by_reset_token(self, credentials: SyftVerifyKey, token: str) -> User:
        return self.get_one(
            credentials=credentials,
            filters={"reset_token": token},
        ).unwrap()

    @as_result(StashException, NotFoundException)
    def get_by_email(self, credentials: SyftVerifyKey, email: str) -> User:
        return self.get_one(
            credentials=credentials,
            filters={"email": email},
        ).unwrap()

    @as_result(StashException)
    def email_exists(self, email: str) -> bool:
        try:
            self.get_by_email(credentials=self.root_verify_key, email=email).unwrap()
            return True
        except NotFoundException:
            return False

    @as_result(StashException)
    def verify_key_exists(self, verify_key: SyftVerifyKey) -> bool:
        try:
            self.get_by_verify_key(
                credentials=self.root_verify_key, verify_key=verify_key
            ).unwrap()
            return True
        except NotFoundException:
            return False

    @as_result(StashException, NotFoundException)
    def get_by_role(self, credentials: SyftVerifyKey, role: ServiceRole) -> User:
        try:
            return self.get_one(
                credentials=credentials,
                filters={"role": role},
            ).unwrap()
        except NotFoundException as exc:
            private_msg = f"User with role {role} not found"
            raise NotFoundException.from_exception(exc, private_message=private_msg)

    @as_result(StashException, NotFoundException)
    def get_by_signing_key(
        self, credentials: SyftVerifyKey, signing_key: SyftSigningKey | str
    ) -> User:
        try:
            return self.get_one(
                credentials=credentials,
                filters={"signing_key": signing_key},
            ).unwrap()
        except NotFoundException as exc:
            private_msg = f"User with signing key {signing_key} not found"
            raise NotFoundException.from_exception(exc, private_message=private_msg)

    @as_result(StashException, NotFoundException)
    def get_by_verify_key(
        self, credentials: SyftVerifyKey, verify_key: SyftVerifyKey
    ) -> User:
        try:
            return self.get_one(
                credentials=credentials,
                filters={"verify_key": verify_key},
            ).unwrap()

        except NotFoundException as exc:
            private_msg = f"User with verify key {verify_key} not found"
            raise NotFoundException.from_exception(exc, private_message=private_msg)



# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...types.result import as_result
from ...types.uid import UID
from .dataset import Dataset


@serializable(canonical_name="DatasetStashSQL", version=1)
class DatasetStash(ObjectStash[Dataset]):
    @as_result(StashException, NotFoundException)
    def get_by_name(self, credentials: SyftVerifyKey, name: str) -> Dataset:
        return self.get_one(credentials=credentials, filters={"name": name}).unwrap()

    @as_result(StashException)
    def search_action_ids(self, credentials: SyftVerifyKey, uid: UID) -> list[Dataset]:
        return self.get_all_active(
            credentials=credentials,
            filters={"action_ids__contains": uid},
        ).unwrap()

    @as_result(StashException)
    def get_all_active(
        self,
        credentials: SyftVerifyKey,
        has_permission: bool = False,
        order_by: str | None = None,
        sort_order: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
        filters: dict | None = None,
    ) -> list[Dataset]:
        # TODO standardize soft delete and move to ObjectStash.get_all
        default_filters = {"to_be_deleted": False}
        filters = filters or {}
        filters.update(default_filters)

        if offset is None:
            offset = 0

        return (
            super()
            .get_all(
                credentials=credentials,
                filters=filters,
                has_permission=has_permission,
                order_by=order_by,
                sort_order=sort_order,
                limit=limit,
                offset=offset,
            )
            .unwrap()
        )
